#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FixedDialogTeacher
from parlai.utils.io import PathManager
from .build import build

import os

EOS_TOKEN = '<eos>'
SELECTION_TOKEN = '<selection>'
YOU_TOKEN = 'YOU:'
THEM_TOKEN = 'THEM:'
SILENCE_TOKEN = '__SILENCE__'

INPUT_TAG = 'input'
DIALOGUE_TAG = 'dialogue'
REFERENTS_TAG = 'referents'
OUTPUT_TAG = 'output'
REAL_IDS_TAG = 'real_ids'

REF_BEGIN_IDX = 0
REF_END_IDX = 1
REF_EOS_IDX = 2
REF_BEGIN_TARGET_IDX = 3

N_OBJECT = 7


def get_tag(tokens, tag):
    """
    Extracts the value inside the given tag.
    """
    start = tokens.index('<' + tag + '>') + 1
    stop = tokens.index('</' + tag + '>')
    return tokens[start:stop]


class OneCommonTeacher(FixedDialogTeacher):
    """
    OneCommon teacher that loads the data from https://github.com/Alab-NII/Reference-
    Resolution.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.datatype = opt['datatype'].split(':')[0]
        build(opt)

        data_path = os.path.join(
            opt['datapath'],
            'onecommon',
            'onecommon-1.0',
            'aaai2020',
            'experiments',
            'data',
            'onecommon',
            self.datatype + '_reference_0.txt',
        )

        if shared and 'episodes' in shared:
            self.episodes = shared['episodes']
        else:
            self._setup_data(data_path)

        self.expected = {}

        self.reset()

    def reset(self):
        self.expected = {}
        super().reset()

    def num_examples(self):
        num_exs = sum(len(episode['dialogue']) // 2 for episode in self.episodes)
        return num_exs

    def num_episodes(self):
        return len(self.episodes)

    def share(self):
        shared = super().share()
        shared['episodes'] = self.episodes
        return shared

    def observe(self, observation):
        if 'metrics' not in observation:
            observation['metrics'] = {}

        # Selection accuracy
        if (
            'output' in observation
            and observation['output'] is not None
            and 'output' in self.expected
            and self.expected['output'] is not None
        ):
            obs_out = int(observation['output'])
            exp_out = int(self.expected['output'])
            observation['metrics']['target_sel_accuracy'] = float(obs_out == exp_out)

        # Reference accuracy
        if (
            'referents' in observation
            and observation['referents'] is not None
            and len(observation['referents']) > 0
            and 'referents' in self.expected
            and self.expected['referents'] is not None
            and len(self.expected['referents']) > 0
        ):
            obs_refs = observation['referents']
            exp_refs = self.expected['referents']
            assert len(obs_refs) == len(exp_refs)

            exact_match = 0.0
            equal_cnt = total_cnt = 0.0
            for obs_ref, exp_ref in zip(obs_refs, exp_refs):
                for obs_tgt, exp_tgt in zip(obs_ref['target'], exp_ref['target']):
                    # Expected sorted referents
                    equal_cnt += float(int(obs_tgt) == int(exp_tgt))
                    total_cnt += 1.0
                if equal_cnt == total_cnt:
                    exact_match += 1.0

            observation['metrics']['referent_exact_match'] = exact_match / len(exp_refs)
            observation['metrics']['referent_accuracy'] = equal_cnt / total_cnt

        return super().observe(observation)

    def get(self, episode_idx, entry_idx):
        episode = self.episodes[episode_idx]
        dialogue = episode['dialogue']
        referents = episode['referents']

        entry_idx *= 2  # Every two utterance is an entry

        action = {}
        action['context'] = episode['context']

        # Fill in teacher's message (THEM)
        sentence = dialogue[entry_idx]
        if sentence is None:
            action['text'] = SILENCE_TOKEN
        else:
            action['text'] = ' '.join(sentence[1:])

        # Fill in learner's response (YOU)
        entry_idx += 1
        sentence = dialogue[entry_idx]
        if sentence is None:
            action['labels'] = [SELECTION_TOKEN]
            action['referents'] = []
        else:
            action['labels'] = [' '.join(sentence[1:])]
            action['referents'] = referents[entry_idx]
        self.expected['referents'] = action['referents']

        # Fill in output at end of dialogue
        if entry_idx < len(dialogue) - 1:
            action['output'] = None
            action['episode_done'] = False
        else:
            action['output'] = episode['output']
            action['episode_done'] = True
            self.expected['output'] = action['output']

        return action

    def _setup_data(self, data_path):
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            raw_data = data_file.readlines()

        self.episodes = []
        for data in raw_data:
            words = data.strip().split()
            context = list(map(float, get_tag(words, INPUT_TAG)))
            dialogue, spans = self._split_dialogue(get_tag(words, DIALOGUE_TAG))
            referents = self._split_referents(get_tag(words, REFERENTS_TAG), spans)
            output = int(get_tag(words, OUTPUT_TAG)[0])
            self.episodes.append(
                {
                    'context': context,
                    'dialogue': dialogue,
                    'referents': referents,
                    'output': output,
                }
            )

    def _split_dialogue(self, words, separator=EOS_TOKEN):
        sentences = []
        spans = []
        start = 0
        for stop in range(len(words)):
            if words[stop] == separator:
                sentences.append(words[start:stop])
                spans.append((start, stop))
                start = stop + 1
        if stop >= start:
            sentences.append(words[start:])
            spans.append((start, len(words) - 1))

        # Dataset contains consecutive turn
        # concatenate utterances for those cases
        dialogue = []
        utterance = sentences[0]
        for i in range(1, len(sentences)):
            if sentences[i - 1][0] == sentences[i][0]:
                utterance += sentences[i][1:]
            else:
                dialogue.append(utterance)
                utterance = sentences[i]
        dialogue.append(utterance)

        if dialogue[0][0] == YOU_TOKEN:
            # Dialogue starts with YOU
            dialogue.insert(0, None)
            spans.insert(0, None)
        if dialogue[-1][0] == THEM_TOKEN:
            # Dialogue starts with THEM
            dialogue.append(None)
            spans.append(None)

        return dialogue, spans

    def _split_referents(self, raw_referents, spans):
        """
        Split the referents.

        The first 3 values are begin idx, end idx, and eos idx The next N_OBJECT values
        are booleans of if the object is referred e.g. 3 4 10 0 1 0 0 0 0 0 means idx 3
        to 4 is a markable of an utterance with <eos> at idx 10, and it refers to the
        2nd dot
        """

        referent_len = 3 + N_OBJECT
        splitted_referents = []
        for i in range(len(raw_referents) // referent_len):
            val = raw_referents[i * referent_len : (i + 1) * referent_len]
            splitted_referents.append(list(map(int, val)))

        referents = []
        idx = 0
        for span in spans:
            if span is None:
                referents.append(None)
                continue

            # span is a (bos index, eos index) of an utterance
            refs = []
            while idx < len(splitted_referents):
                if splitted_referents[idx][REF_EOS_IDX] == span[1]:
                    ref = {
                        'begin': splitted_referents[idx][REF_BEGIN_IDX] - (span[0] + 1),
                        'end': splitted_referents[idx][REF_END_IDX] - (span[0] + 1),
                        'target': splitted_referents[idx][REF_BEGIN_TARGET_IDX:],
                    }
                    refs.append(ref)
                    idx += 1
                else:
                    break
            referents.append(refs)

        return referents


class DefaultTeacher(OneCommonTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
from typing import Dict

from omegaconf import DictConfig
import parlai.utils.logging as logging
from parlai.core.agents import create_agent
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
from parlai.crowdsourcing.tasks.model_chat.constants import AGENT_1
from parlai.utils.strings import normalize_reply


class TurkLikeAgent:
    """
    Will act like a Turker but actually contains a bot agent.
    """

    def __init__(self, opt, model_name, model_agent, num_turns, semaphore=None):
        self.opt = opt
        self.model_agent = model_agent
        self.id = AGENT_1
        self.num_turns = num_turns
        self.turn_idx = 0
        self.semaphore = semaphore
        self.worker_id = model_name
        self.hit_id = 'none'
        self.assignment_id = 'none'
        self.some_agent_disconnected = False
        self.hit_is_abandoned = False
        self.hit_is_returned = False
        self.disconnected = False
        self.hit_is_expired = False

    def act(self, timeout=None):
        _ = timeout  # The model doesn't care about the timeout
        if self.semaphore:
            with self.semaphore:
                act_out = self.model_agent.act()
        else:
            act_out = self.model_agent.act()
        act_out = Message(act_out)
        # Wrap as a Message for compatibility with older ParlAI models

        if 'dict_lower' in self.opt and not self.opt['dict_lower']:
            # model is cased so we don't want to normalize the reply like below
            final_message_text = act_out['text']
        else:
            final_message_text = normalize_reply(act_out['text'])

        act_out.force_set('text', final_message_text)
        assert ('episode_done' not in act_out) or (not act_out['episode_done'])
        self.turn_idx += 1
        return {**act_out, 'episode_done': False}

    def observe(self, observation, increment_turn: bool = True):
        """
        Need to protect the observe also with a semaphore for composed models where an
        act() may be called within an observe()
        """
        logging.info(
            f'{self.__class__.__name__}: In observe() before semaphore, self.turn_idx is {self.turn_idx} and observation is {observation}'
        )
        new_ob = copy.deepcopy(observation)
        if self.semaphore:
            with self.semaphore:
                self.model_agent.observe(new_ob)
        else:
            self.model_agent.observe(new_ob)
        logging.info(
            f'{self.__class__.__name__}: In observe() AFTER semaphore, self.turn_idx: {self.turn_idx}, observation["text"]: {new_ob["text"]}'
        )

        if increment_turn:
            self.turn_idx += 1

    def shutdown(self):
        pass

    def reset(self):
        self.model_agent.reset()

    @staticmethod
    def get_bot_agents(
        args: DictConfig, model_opts: Dict[str, str], no_cuda=False
    ) -> Dict[str, dict]:
        """
        Return shared bot agents.

        Pass in model opts with the `model_opts` arg, where `model_opts` is a dictionary
        whose keys are model names and whose values are strings that specify model
        params (i.e. `--model image_seq2seq`).
        """

        # Set up overrides
        model_overrides = {'model_parallel': args.blueprint.task_model_parallel}
        if no_cuda:
            # If we load many models at once, we have to keep it on CPU
            model_overrides['no_cuda'] = no_cuda
        else:
            logging.warn(
                'WARNING: MTurk task has no_cuda FALSE. Models will run on GPU. Will '
                'not work if loading many models at once.'
            )

        # Convert opt strings to Opt objects
        parser = ParlaiParser(True, True)
        parser.set_params(**model_overrides)
        processed_opts = {}
        for name, opt_string in model_opts.items():
            processed_opts[name] = parser.parse_args(opt_string.split())

        # Load and share all model agents
        logging.info(
            f'Got {len(list(processed_opts.keys()))} models: {processed_opts.keys()}.'
        )
        shared_bot_agents = {}
        for model_name, model_opt in processed_opts.items():
            logging.info('\n\n--------------------------------')
            logging.info(f'model_name: {model_name}, opt_dict: {model_opt}')
            model_agent = create_agent(model_opt, requireModelExists=True)
            shared_bot_agents[model_name] = model_agent.share()
        return shared_bot_agents



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Simple IR baselines.

We plan to implement the following variants:
Given an input message, either:
(i) find the most similar message in the (training) dataset and output the
  response from that exchange; or
(ii) find the most similar response to the input directly.
(iii) if label_candidates are provided, simply ranks them according to their
  similarity to the input message.

Currently only (iii) is used.

Additonally, TFIDF is either used (requires building a dictionary) or not,
depending on whether you train on the train set first, or not.
"""

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import math
from collections.abc import Sequence
import heapq
import json

import parlai.utils.torch as torch_utils
from parlai.core.agents import Agent
from parlai.core.dict import DictionaryAgent
from parlai.utils.io import PathManager


class MaxPriorityQueue(Sequence):
    """
    Fixed-size priority queue keeping the max_size largest items.
    """

    def __init__(self, max_size):
        """
        Initialize priority queue.

        :param max_size: maximum capacity of priority queue.
        """
        self.capacity = max_size
        self.lst = []

    def add(self, item, priority):
        """
        Add element to the queue, with a separate priority if desired.

        Element will not be added if the queue is at capacity and the element
        has lower priority than the lowest currently in the queue.

        :param item:
            item to add to queue.
        :param priority:
            priority to use for item.
        """
        if len(self.lst) < self.capacity:
            heapq.heappush(self.lst, (priority, item))
        elif priority > self.lst[0][0]:
            heapq.heapreplace(self.lst, (priority, item))

    def __getitem__(self, key):
        """
        Get item at specified index.

        :param key: integer index into priority queue, 0 <= index < max_size.

        :returns: item stored at the specified index.
        """
        return sorted(self.lst)[key][1]

    def __len__(self):
        """
        Return length of priority queue.
        """
        return len(self.lst)


stopwords = {
    'i',
    'a',
    'an',
    'are',
    'about',
    'as',
    'at',
    'be',
    'by',
    'for',
    'from',
    'how',
    'in',
    'is',
    'it',
    'of',
    'on',
    'or',
    'that',
    'the',
    'this',
    'to',
    'was',
    'what',
    'when',
    'where',
    '--',
    '?',
    '.',
    "''",
    "''",
    "``",
    ',',
    'do',
    'see',
    'want',
    'people',
    'and',
    "n't",
    "me",
    'too',
    'own',
    'their',
    '*',
    "'s",
    'not',
    'than',
    'other',
    'you',
    'your',
    'know',
    'just',
    'but',
    'does',
    'really',
    'have',
    'into',
    'more',
    'also',
    'has',
    'any',
    'why',
    'will',
}


def score_match(query_rep, text, length_penalty, dictionary=None):
    """
    Calculate the score match between the query representation the text.

    :param query_rep:
        base query representation to match text again.
    :param text:
        string to compare against query_rep for matching tokens
    :param length_penalty:
        scores are divided by the norm taken to this power
    :param dictionary:
        optional dictionary to use to tokenize text

    :returns:
        float score of match
    """
    if text == "":
        return 0
    words = [w for w in dictionary.tokenize(text.lower())]
    score = 0
    rw = query_rep['words']
    used = {}
    for w in words:
        if w in rw and w not in used:
            score += rw[w]
        used[w] = True
    norm = math.sqrt(len(used))
    norm = math.pow(norm * query_rep['norm'], length_penalty)
    if norm > 1:
        score /= norm
    return score


def rank_candidates(query_rep, cands, length_penalty, dictionary):
    """
    Rank candidates given representation of query.

    :param query_rep:
        base query representation to match text again.
    :param cands:
        strings to compare against query_rep for matching tokens
    :param length_penalty:
        scores are divided by the norm taken to this power
    :dictionary:
        dictionary to use to tokenize text

    :returns:
        ordered list of candidate strings in score-ranked order
    """
    if True:
        mpq = MaxPriorityQueue(100)
        for c in cands:
            score = score_match(query_rep, c, length_penalty, dictionary)
            mpq.add(c, score)
        return list(reversed(mpq))
    else:
        cands = list(cands)
        score = [0] * len(cands)
        for i, c in enumerate(cands):
            score[i] = -score_match(query_rep, c, length_penalty, dictionary)
        r = [i[0] for i in sorted(enumerate(score), key=lambda x: x[1])]
        res = []
        for i in range(min(100, len(score))):
            res.append(cands[r[i]])
        return res


class IrBaselineAgent(Agent):
    """
    Information Retrieval baseline.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command line args specific to this agent.
        """
        parser = parser.add_argument_group('IrBaseline Arguments')
        parser.add_argument(
            '-lp',
            '--length_penalty',
            type=float,
            default=0.5,
            help='length penalty for responses',
        )
        parser.add_argument(
            '-hsz',
            '--history_size',
            type=int,
            default=1,
            help='number of utterances from the dialogue history to take use '
            'as the query',
        )
        parser.add_argument(
            '--label_candidates_file',
            type=str,
            default=None,
            help='file of candidate responses to choose from',
        )
        cls.dictionary_class().add_cmdline_args(parser, partial_opt=partial_opt)
        return parser

    @classmethod
    def dictionary_class(cls):
        return DictionaryAgent

    def __init__(self, opt, shared=None):
        """
        Initialize agent.
        """
        super().__init__(opt)
        self.id = 'IRBaselineAgent'
        self.length_penalty = float(opt['length_penalty'])
        self.dictionary = DictionaryAgent(opt)
        self.opt = opt
        self.history = []
        self.episodeDone = True
        if opt.get('label_candidates_file'):
            f = open(opt.get('label_candidates_file'))
            self.label_candidates = f.read().split('\n')

    def reset(self):
        """
        Reset agent properties.
        """
        self.observation = None
        self.history = []
        self.episodeDone = True

    def observe(self, obs):
        """
        Store and remember incoming observation message dict.
        """
        self.observation = obs
        self.dictionary.observe(obs)
        if self.episodeDone:
            self.history = []
        if 'text' in obs:
            self.history.append(obs.get('text', ''))
        self.episodeDone = obs.get('episode_done', False)
        return obs

    def act(self):
        """
        Generate a response to the previously seen observation(s).
        """
        if self.opt.get('datatype', '').startswith('train'):
            self.dictionary.act()

        obs = self.observation
        reply = {}
        reply['id'] = self.getID()

        # Rank candidates
        cands = None
        if obs.get('label_candidates', False) and len(obs['label_candidates']) > 0:
            cands = obs['label_candidates']
        if hasattr(self, 'label_candidates'):
            # override label candidates with candidate file if set
            cands = self.label_candidates
        if cands:
            hist_sz = self.opt.get('history_size', 1)
            left_idx = max(0, len(self.history) - hist_sz)
            text = ' '.join(self.history[left_idx : len(self.history)])
            rep = self.build_query_representation(text)
            reply['text_candidates'] = rank_candidates(
                rep, cands, self.length_penalty, self.dictionary
            )
            reply['text'] = reply['text_candidates'][0]
        return reply

    def save(self, path=None):
        """
        Save dictionary tokenizer if available.
        """
        path = self.opt.get('model_file', None) if path is None else path
        if path:
            self.dictionary.save(path + '.dict')
            data = {}
            data['opt'] = self.opt
            torch_utils.atomic_save(data, path)
            with PathManager.open(path + '.opt', 'w') as handle:
                json.dump(self.opt, handle)

    def build_query_representation(self, query):
        """
        Build representation of query, e.g. words or n-grams.

        :param query: string to represent.

        :returns: dictionary containing 'words' dictionary (token => frequency)
                  and 'norm' float (square root of the number of tokens)
        """
        rep = {}
        rep['words'] = {}
        words = [w for w in self.dictionary.tokenize(query.lower())]
        rw = rep['words']
        used = {}
        for w in words:
            assert len(self.dictionary.freq) > 0
            rw[w] = 1.0 / (1.0 + math.log(1.0 + self.dictionary.freq[w]))
            used[w] = True
        rep['norm'] = math.sqrt(len(words))
        return rep



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# py parlai/chat_service/tasks/overworld_demo/run.py --debug --verbose

from parlai.core.worlds import World
from parlai.chat_service.services.messenger.worlds import OnboardWorld
from parlai.core.agents import create_agent_from_shared


# ---------- Chatbot demo ---------- #
class MessengerBotChatOnboardWorld(OnboardWorld):
    """
    Example messenger onboarding world for Chatbot Model.
    """

    @staticmethod
    def generate_world(opt, agents):
        return MessengerBotChatOnboardWorld(opt=opt, agent=agents[0])

    def parley(self):
        self.episodeDone = True


class MessengerBotChatTaskWorld(World):
    """
    Example one person world that talks to a provided agent (bot).
    """

    MAX_AGENTS = 1
    MODEL_KEY = 'blender_90M'

    def __init__(self, opt, agent, bot):
        self.agent = agent
        self.episodeDone = False
        self.model = bot
        self.first_time = True

    @staticmethod
    def generate_world(opt, agents):
        if opt['models'] is None:
            raise RuntimeError("Model must be specified")
        return MessengerBotChatTaskWorld(
            opt,
            agents[0],
            create_agent_from_shared(
                opt['shared_bot_params'][MessengerBotChatTaskWorld.MODEL_KEY]
            ),
        )

    @staticmethod
    def assign_roles(agents):
        agents[0].disp_id = 'ChatbotAgent'

    def parley(self):
        if self.first_time:
            self.agent.observe(
                {
                    'id': 'World',
                    'text': 'Welcome to the ParlAI Chatbot demo. '
                    'You are now paired with a bot - feel free to send a message.'
                    'Type [DONE] to finish the chat, or [RESET] to reset the dialogue history.',
                }
            )
            self.first_time = False
        a = self.agent.act()
        if a is not None:
            if '[DONE]' in a['text']:
                self.episodeDone = True
            elif '[RESET]' in a['text']:
                self.model.reset()
                self.agent.observe({"text": "[History Cleared]", "episode_done": False})
            else:
                print("===act====")
                print(a)
                print("~~~~~~~~~~~")
                self.model.observe(a)
                response = self.model.act()
                print("===response====")
                print(response)
                print("~~~~~~~~~~~")
                self.agent.observe(response)

    def episode_done(self):
        return self.episodeDone

    def shutdown(self):
        self.agent.shutdown()


# ---------- Overworld -------- #
class MessengerOverworld(World):
    """
    World to handle moving agents to their proper places.
    """

    def __init__(self, opt, agent):
        self.agent = agent
        self.opt = opt
        self.first_time = True
        self.episodeDone = False

    @staticmethod
    def generate_world(opt, agents):
        return MessengerOverworld(opt, agents[0])

    @staticmethod
    def assign_roles(agents):
        for a in agents:
            a.disp_id = 'Agent'

    def episode_done(self):
        return self.episodeDone

    def parley(self):
        if self.first_time:
            self.agent.observe(
                {
                    'id': 'Overworld',
                    'text': 'Welcome to the overworld for the ParlAI messenger '
                    'chatbot demo. Please type "begin" to start, or "exit" to exit',
                    'quick_replies': ['begin', 'exit'],
                }
            )
            self.first_time = False
        a = self.agent.act()
        if a is not None and a['text'].lower() == 'exit':
            self.episode_done = True
            return 'EXIT'
        if a is not None and a['text'].lower() == 'begin':
            self.episodeDone = True
            return 'default'
        elif a is not None:
            self.agent.observe(
                {
                    'id': 'Overworld',
                    'text': 'Invalid option. Please type "begin".',
                    'quick_replies': ['begin'],
                }
            )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import sys

if sys.version_info < (3, 0):
    raise RuntimeError('ParlAI requires Python 3.')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


# ------------------------------------------------------------------------------
# Modules
# ------------------------------------------------------------------------------


class StackedBRNN(nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers,
        dropout_rate=0,
        dropout_output=False,
        rnn_type=nn.LSTM,
        concat_layers=False,
        padding=False,
    ):
        super(StackedBRNN, self).__init__()
        self.padding = padding
        self.dropout_output = dropout_output
        self.dropout_rate = dropout_rate
        self.num_layers = num_layers
        self.concat_layers = concat_layers
        self.rnns = nn.ModuleList()
        for i in range(num_layers):
            input_size = input_size if i == 0 else 2 * hidden_size
            self.rnns.append(
                rnn_type(input_size, hidden_size, num_layers=1, bidirectional=True)
            )

    def forward(self, x, x_mask):
        """
        Can choose to either handle or ignore variable length sequences.

        Always handle padding in eval.
        """
        # No padding necessary.
        if x_mask.data.sum() == 0:
            return self._forward_unpadded(x, x_mask)
        # Pad if we care or if its during eval.
        if self.padding or not self.training:
            return self._forward_padded(x, x_mask)
        # We don't care.
        return self._forward_unpadded(x, x_mask)

    def _forward_unpadded(self, x, x_mask):
        """
        Faster encoding that ignores any padding.
        """
        # Transpose batch and sequence dims
        x = x.transpose(0, 1).contiguous()

        # Encode all layers
        outputs = [x]
        for i in range(self.num_layers):
            rnn_input = outputs[-1]

            # Apply dropout to hidden input
            if self.dropout_rate > 0:
                rnn_input = F.dropout(
                    rnn_input, p=self.dropout_rate, training=self.training
                )
            # Forward
            rnn_output = self.rnns[i](rnn_input)[0]
            outputs.append(rnn_output)

        # Concat hidden layers
        if self.concat_layers:
            output = torch.cat(outputs[1:], 2)
        else:
            output = outputs[-1]

        # Transpose back
        output = output.transpose(0, 1).contiguous()

        # Dropout on output layer
        if self.dropout_output and self.dropout_rate > 0:
            output = F.dropout(output, p=self.dropout_rate, training=self.training)
        return output

    def _forward_padded(self, x, x_mask):
        """
        Slower (significantly), but more precise, encoding that handles padding.
        """
        # Compute sorted sequence lengths
        lengths = x_mask.data.eq(0).long().sum(1).squeeze()
        _, idx_sort = torch.sort(lengths, dim=0, descending=True)
        _, idx_unsort = torch.sort(idx_sort, dim=0)

        lengths = list(lengths[idx_sort])
        idx_sort = Variable(idx_sort)
        idx_unsort = Variable(idx_unsort)

        # Sort x
        x = x.index_select(0, idx_sort)

        # Transpose batch and sequence dims
        x = x.transpose(0, 1).contiguous()

        # Pack it up
        rnn_input = nn.utils.rnn.pack_padded_sequence(x, lengths)

        # Encode all layers
        outputs = [rnn_input]
        for i in range(self.num_layers):
            rnn_input = outputs[-1]

            # Apply dropout to input
            if self.dropout_rate > 0:
                dropout_input = F.dropout(
                    rnn_input.data, p=self.dropout_rate, training=self.training
                )
                rnn_input = nn.utils.rnn.PackedSequence(
                    dropout_input, rnn_input.batch_sizes
                )
            outputs.append(self.rnns[i](rnn_input)[0])

        # Unpack everything
        for i, o in enumerate(outputs[1:], 1):
            outputs[i] = nn.utils.rnn.pad_packed_sequence(o)[0]

        # Concat hidden layers or take final
        if self.concat_layers:
            output = torch.cat(outputs[1:], 2)
        else:
            output = outputs[-1]

        # Transpose and unsort
        output = output.transpose(0, 1).contiguous()
        output = output.index_select(0, idx_unsort)

        # Dropout on output layer
        if self.dropout_output and self.dropout_rate > 0:
            output = F.dropout(output, p=self.dropout_rate, training=self.training)
        return output


class SeqAttnMatch(nn.Module):
    """
    Given sequences X and Y, match sequence Y to each element in X.

    * o_i = sum(alpha_j * y_j) for i in X
    * alpha_j = softmax(y_j * x_i)
    """

    def __init__(self, input_size, identity=False):
        super(SeqAttnMatch, self).__init__()
        if not identity:
            self.linear = nn.Linear(input_size, input_size)
        else:
            self.linear = None

    def forward(self, x, y, y_mask):
        """
        Input shapes:

            x = batch * len1 * h
            y = batch * len2 * h
            y_mask = batch * len2
        Output shapes:
            matched_seq = batch * len1 * h
        """
        # Project vectors
        if self.linear:
            x_proj = self.linear(x.view(-1, x.size(2))).view(x.size())
            x_proj = F.relu(x_proj)
            y_proj = self.linear(y.view(-1, y.size(2))).view(y.size())
            y_proj = F.relu(y_proj)
        else:
            x_proj = x
            y_proj = y

        # Compute scores
        scores = x_proj.bmm(y_proj.transpose(2, 1))

        # Mask padding
        y_mask = y_mask.unsqueeze(1).expand(scores.size())
        scores.data.masked_fill_(y_mask.data, -float('inf'))

        # Normalize with softmax
        alpha_flat = F.softmax(scores.view(-1, y.size(1)), dim=-1)
        alpha = alpha_flat.view(-1, x.size(1), y.size(1))

        # Take weighted average
        matched_seq = alpha.bmm(y)
        return matched_seq


class BilinearSeqAttn(nn.Module):
    """
    A bilinear attention layer over a sequence X w.r.t y:

    * o_i = softmax(x_i'Wy) for x_i in X.

    Optionally don't normalize output weights.
    """

    def __init__(self, x_size, y_size, identity=False):
        super(BilinearSeqAttn, self).__init__()
        if not identity:
            self.linear = nn.Linear(y_size, x_size)
        else:
            self.linear = None

    def forward(self, x, y, x_mask):
        """
        x = batch * len * h1
        y = batch * h2
        x_mask = batch * len
        """
        Wy = self.linear(y) if self.linear is not None else y
        xWy = x.bmm(Wy.unsqueeze(2)).squeeze(2)
        xWy.data.masked_fill_(x_mask.data, -float('inf'))
        self.xWy = xWy
        if self.training:
            # In training we output log-softmax for NLL
            alpha = F.log_softmax(xWy, dim=-1)
        else:
            # ...Otherwise 0-1 probabilities
            # alpha = F.softmax(xWy)

            # Note: We found better eval performance with unnormalized weights
            #       here
            alpha = xWy.exp()
        return alpha


class LinearSeqAttn(nn.Module):
    """
    Self attention over a sequence:

    * o_i = softmax(Wx_i) for x_i in X.
    """

    def __init__(self, input_size):
        super(LinearSeqAttn, self).__init__()
        self.linear = nn.Linear(input_size, 1)

    def forward(self, x, x_mask):
        """
        x = batch * len * hdim
        x_mask = batch * len
        """
        x_flat = x.view(-1, x.size(-1))
        scores = self.linear(x_flat).view(x.size(0), x.size(1))
        scores.data.masked_fill_(x_mask.data, -float('inf'))
        alpha = F.softmax(scores, dim=-1)
        return alpha


# ------------------------------------------------------------------------------
# Functional
# ------------------------------------------------------------------------------


def uniform_weights(x, x_mask):
    """
    Return uniform weights over non-masked input.
    """
    alpha = Variable(torch.ones(x.size(0), x.size(1)))
    if x.data.is_cuda:
        alpha = alpha.cuda()
    alpha = alpha * x_mask.eq(0).float()
    alpha = alpha / alpha.sum(1).expand(alpha.size())
    return alpha


def weighted_avg(x, weights):
    """x = batch * len * d
    weights = batch * len
    """
    return weights.unsqueeze(1).bmm(x).squeeze(1)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Dict, Any, TYPE_CHECKING
from mephisto.abstractions.blueprints.abstract.static_task.static_agent_state import (
    StaticAgentState,
)
import time

if TYPE_CHECKING:
    from mephisto.data_model.packet import Packet


DATA_FILE = "agent_data.json"


class AcuteEvalAgentState(StaticAgentState):
    """
    Agent state for acute eval tasks.

    Equivalent to StaticAgentState but doesn't have file IO.
    """

    def get_parsed_data(self) -> List[Dict[str, Any]]:
        data = self.get_data()
        assert data is not None, "Should only check parsed data for completed tasks"
        response_list = []
        inputs: List[Dict[str, Any]] = data["inputs"]
        outputs = data["outputs"]
        assert inputs is not None
        assert outputs is not None
        for idx in range(len(inputs)):
            entry: Dict[str, Any] = {}
            entry.update(inputs[idx])
            entry.update(outputs["final_data"][idx])
            response_list.append(entry)
        return response_list

    def update_data(self, packet: "Packet") -> None:
        """
        Process the incoming data packet, and handle updating the state.
        """
        assert (
            packet.data.get("MEPHISTO_is_submit") is True
        ), "Static tasks should only have final act"
        self.state["times"]["task_end"] = time.time()
        self.state["outputs"] = packet.data["task_data"]
        self.save_data()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.teachers import ChunkTeacher
from parlai.tasks.wikipedia.build import build
from parlai.utils.misc import warn_once
import parlai.tasks.md_gender.utils as gend_utils

from copy import deepcopy
import json
import os
import random
import re
import spacy
from typing import List


NLP = spacy.load('en_core_web_sm')

DEBUG = False


def check_if_person(text):
    """
    Using spacy, check if the title of the Wikipedia passage is a person.
    """
    doc = NLP(text)
    is_person = False
    ents = [ent for ent in doc.ents]
    for ent in ents:
        if ent.label_ == 'PERSON':
            is_person = True
    return is_person


def get_gender(text):
    """
    Determine gender by the count of referring pronouns in the biography (he, she,
    they).
    """
    he_count = len(re.findall(' he ', text.lower()))
    she_count = len(re.findall(' she ', text.lower()))
    they_count = len(re.findall(' they ', text.lower()))

    if he_count == max(he_count, she_count, they_count):
        return gend_utils.MASC
    elif she_count == max(he_count, she_count, they_count):
        return gend_utils.FEM
    else:
        nonbinary_count = len(re.findall(' non-binary ', text.lower()))
        if nonbinary_count > 0:
            return gend_utils.NONBINARY
        return gend_utils.NEUTRAL


class WikipediaTeacher(ChunkTeacher):
    """
    Wikipedia gender.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = gend_utils.add_common_args(parser)
        agent = parser.add_argument_group('Wiki gender')
        agent.add_argument(
            '--class-task',
            type=str,
            default='all',
            choices=['single', 'all'],
            help='Rank against all possibilities vs. F/M/N/NB',
        )
        agent.add_argument(
            '--mask-gendered-words',
            type='bool',
            default=False,
            help='Mask explicitly gendered words.',
        )
        return parser

    def __init__(self, opt, shared=None):
        if shared is None:
            # set map
            self.opt = opt
            self._set_chunk_idx_to_file(opt)
        else:
            self.chunk_idx_to_file = shared['chunk_idx_to_file']

        self.class_task = opt['class_task']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = (
            opt['add_unknown_classes'] and self.is_train and self.class_task == 'all'
        )

        if self.class_task == 'all':
            self.label_candidates = gend_utils.ALL_CANDS
        elif self.class_task == 'single':
            self.label_candidates = {
                'about': [
                    gend_utils.FEM,
                    gend_utils.MASC,
                    gend_utils.NEUTRAL,
                    gend_utils.NONBINARY,
                ]
            }

        self.mask_gendered_words = opt['mask_gendered_words']
        if self.mask_gendered_words:
            male, female = gend_utils.get_explicitly_gendered_words(self.opt)
            self.gendered_list = male + female

        self.counts = {
            gend_utils.FEM: 0,
            gend_utils.MASC: 0,
            gend_utils.NEUTRAL: 0,
            gend_utils.NONBINARY: 0,
        }
        super().__init__(opt, shared)

    def _set_chunk_idx_to_file(self, opt):
        # download wikipedia data
        wiki_opt = deepcopy(opt)
        wiki_opt['task'] = 'wikipedia:full'
        build(wiki_opt)

        # now divide data into subfolders
        data_folder = self._get_data_folder()
        self.chunk_idx_to_file = {}
        i = 0
        for subdir in os.listdir(data_folder):
            if subdir != 'README.md':
                for fle in os.listdir(os.path.join(data_folder, subdir)):
                    self.chunk_idx_to_file[i] = os.path.join(data_folder, subdir, fle)
                    i += 1

    def _get_data_folder(self):
        return os.path.join(self.opt['datapath'], 'wikipedia/full/wiki_full_extracted')

    def get_num_samples(self, opt) -> int:
        """
        Return the number of samples given the datatype.
        """
        datatype = opt['datatype']
        if 'train' in datatype:
            return 12774693, 12774693
        elif 'valid' in datatype:
            return 7410, 7410
        else:
            return 7441, 7441

    def get_fold_chunks(self, opt) -> List[int]:  # type: ignore
        """
        Return a list of chunk IDs (integer).

        Given the datatype (train/test/valid), return the list of chunk IDs that
        correspond to that split.
        """
        datatype = opt['datatype']
        all_chunk_idxs = list(self.chunk_idx_to_file.keys())
        if DEBUG:
            print(f'Total chunks: {len(all_chunk_idxs)}')
        if 'train' in datatype:
            return all_chunk_idxs[:-10]
        elif 'valid' in datatype:
            return all_chunk_idxs[-10:-5]
        else:
            return all_chunk_idxs[-5:]

    def load_from_chunk(self, chunk_idx: int):
        """
        [Abstract] Given the chunk index, load examples from that chunk.

        Return a list of tuples. The function `_create_message` will take these tuples
        to form the Message object that is returned by the teacher.
        """
        output = []
        chunk_path = self.chunk_idx_to_file[chunk_idx]

        extra_data = []
        with open(chunk_path) as wf:
            for article_json in wf:
                article = json.loads(article_json)
                title = article['title']
                text = article['text']

                title = title.split(' (')[0]
                is_person = check_if_person(title)
                if not is_person:
                    continue

                gender = get_gender(text)

                label = f'ABOUT:{gender}'
                for par in text.split('\n'):
                    if par:
                        output.append((par, title, label, gender, 'about'))
                        self.counts[gender] += 1

                        if self.add_unknown_classes:
                            extra_data.append(
                                (
                                    par,
                                    title,
                                    f'SELF:{gend_utils.UNKNOWN}',
                                    gender,
                                    'self',
                                )
                            )
                            extra_data.append(
                                (
                                    par,
                                    title,
                                    f'PARTNER:{gend_utils.NEUTRAL}',
                                    gender,
                                    'partner',
                                )
                            )

        if len(extra_data) > 0:
            # possibly sample unknown classes
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(extra_data))
                sampled = random.sample(extra_data, to_samp)
                output += sampled
            else:
                output += extra_data

        if DEBUG:
            print('\n\nGender count update:')
            for k, v in self.counts.items():
                print(f'{k}: {v}')

        if (self.is_train and self.opt['balance']) or (
            self.is_valid and self.opt['balance_valid']
        ):
            exclude_lst = [
                f'ABOUT:{gend_utils.NONBINARY}',
                f'SELF:{gend_utils.UNKNOWN}',
                f'PARTNER:{gend_utils.NEUTRAL}',
            ]  # not enough of each of these examples to balance
            output = gend_utils.balance_data(output, key=2, exclude_labels=exclude_lst)

        if len(output) == 0:
            warn_once(f'CHUNK {chunk_idx} is empty')

        return output

    def _mask_gendered_words(self, text):
        return gend_utils.mask_gendered_words(text, self.gendered_list)

    def create_message(self, queue_output, entry_idx=0) -> 'Message':
        """
        [Abstract] Given the tuple output of the queue, return an act.
        """
        par, title, lbl, gender, class_type = queue_output
        if self.class_task == 'all':
            if class_type == 'self':
                labels = gend_utils.UNKNOWN_LABELS['self']  # Not True neutral
            else:
                labels = [lbl]
        elif self.class_task == 'single':
            labels = [gender]

        if self.mask_gendered_words:
            par = self._mask_gendered_words(par)

        return Message(
            {
                'text': par,
                'name': title,
                'labels': labels,
                'label_candidates': self.label_candidates[class_type],
                'episode_done': True,
                'id': 'Wikipedia Gender',
            }
        )

    def share(self):
        shared = super().share()
        shared['chunk_idx_to_file'] = self.chunk_idx_to_file
        return shared


def write_gender_to_file(open_file, loaded_data):
    prev = None
    for _, title, _, gender, _ in loaded_data:
        if title != prev:
            open_file.write(f'{title}\t{gender}\n')
            prev = title



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Teachers for the MovieDialog task.

From Dodge et al. '15. Link: https://arxiv.org/abs/1511.06931

Task 1: Closed-domain QA dataset asking templated questions about movies,
answerable from Wikipedia.

Task 2: Questions asking for movie recommendations.

Task 3: Dialogs discussing questions about movies as well as recommendations.

Task 4: Dialogs discussing Movies from Reddit (the /r/movies SubReddit).
"""
from parlai.core.teachers import FbDeprecatedDialogTeacher, MultiTaskTeacher
from .build import build

import copy
import os

tasks = {}
tasks[1] = os.path.join('task1_qa', 'task1_qa_pipe_')
tasks[2] = os.path.join('task2_recs', 'task2_recs_')
tasks[3] = os.path.join('task3_qarecs', 'task3_qarecs_pipe_')
tasks[4] = os.path.join('task4_reddit', 'task4_reddit', 'task4_reddit_pipeless_')


def _path(task, opt):
    # Build the data if it doesn't exist.
    build(opt)
    suffix = ''
    dt = opt['datatype'].split(':')[0]
    if dt == 'train':
        suffix = 'train'
    elif dt == 'test':
        suffix = 'test'
    elif dt == 'valid':
        suffix = 'dev'

    datafile = os.path.join(
        opt['datapath'],
        'MovieDialog',
        'movie_dialog_dataset',
        '{t}{s}.txt'.format(t=tasks[int(task)], s=suffix),
    )
    if int(task) == 4:
        if dt == 'train':
            candpath = None
        else:
            candpath = datafile.replace(suffix + '.txt', 'cand-{dt}.txt'.format(dt=dt))
    else:
        candpath = os.path.join(
            opt['datapath'], 'MovieDialog', 'movie_dialog_dataset', 'entities.txt'
        )
    return datafile, candpath


# The knowledge base of facts that can be used to answer questions.
class KBTeacher(FbDeprecatedDialogTeacher):
    """
    Simple text entry with each movie's facts in the knowledge base.
    """

    def __init__(self, opt, shared=None):
        """
        Initialize teacher.
        """
        build(opt)
        opt['datafile'] = os.path.join(
            opt['datapath'], 'MovieDialog', 'movie_dialog_dataset', 'movie_kb.txt'
        )
        super().__init__(opt, shared)


# Single task.
class TaskTeacher(FbDeprecatedDialogTeacher):
    """
    Teacher with single task, specified by moviedialog:task:N.
    """

    def __init__(self, opt, shared=None):
        """
        Initialize teacher.
        """
        try:
            # expecting "moviedialog:task:N"
            self.task = opt['task'].split(':')[2]
        except IndexError:
            self.task = '1'  # default task
        opt['datafile'], opt['cands_datafile'] = _path(self.task, opt)
        super().__init__(opt, shared)


# By default train on all tasks at once.
class DefaultTeacher(MultiTaskTeacher):
    """
    By default will load teacher with all four tasks.
    """

    def __init__(self, opt, shared=None):
        """
        Initialize teacher.
        """
        opt = copy.deepcopy(opt)
        opt['task'] = ','.join(
            'moviedialog:Task:%d' % (i + 1) for i in range(len(tasks))
        )
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager
from .build import build
import os


class QAZRETeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        # store datatype
        self.dt = opt['datatype'].split(':')[0]

        # store identifier for the teacher in the dialog
        self.id = 'qazre'

        # build the data
        build(opt)

        # set the download path
        opt['datafile'] = os.path.join(opt['datapath'], 'QA-ZRE', 'relation_splits')

        super().__init__(opt, shared)

    def setup_data(self, input_path):

        print('loading: ' + input_path)

        new_episode = True

        # function to parse the episodes
        def extract_qa(qa_data):
            line_data = qa_data.split('\t')
            question_type, anon_question, deanon, context = line_data[:4]
            answer = line_data[4:]
            if answer == []:
                answer = ['No answer']
            return context + '\n' + anon_question.replace('XXX', deanon), answer

        # parse and output the episodes
        for fname in os.listdir(input_path):
            if fname.startswith(self.dt):
                with PathManager.open(os.path.join(input_path, fname)) as file:
                    file_data = file.read().split('\n')[:-1]
                for line in file_data:
                    question, answer = extract_qa(line)

                    yield (question, answer, None, None), new_episode


class DefaultTeacher(QAZRETeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from argparse import ArgumentParser
import json

from projects.self_feeding.utils import (
    Parley,
    extract_parlai_episodes,
    add_person_tokens,
)
from parlai.utils.io import PathManager


def setup_args():
    parser = ArgumentParser()
    parser.add_argument('-if', '--infile', type=str)
    parser.add_argument('-of', '--outfile', type=str)
    config = vars(parser.parse_args())

    return config


def main(config):
    """
    Creates .unfiltered files from .sliced files.

    input: a .sliced file of logs (in ParlaiDialog format) from Mturk task 1, each of
        which starts with an initial prompt or topic request, and ends with a y_exp
    output: a .unfiltered file (in self-feeding format) with every utterance output by
        bot used as a label (i.e., act as though the bot was a human and we want to
        train in a normal supervised way).
    """
    examples = []
    episodes = [e for e in extract_parlai_episodes(config['infile'])]
    for episode in episodes:
        history = []
        num_parleys = len(episode)
        for i, parley in enumerate(episode):
            if i == 0:  # Don't include the topic request
                history.append(parley.response)
                continue
            elif i == num_parleys - 1:
                # ultimate turn was correction request and explanation
                continue
            else:
                example = Parley(
                    context=add_person_tokens(history, last_speaker=1),
                    response=parley.context,  # What the bot said
                )
                examples.append(example)
                history.append(parley.context)
                history.append(parley.response)

    # Write new episodes to self-feeding format
    with PathManager.open(config['outfile'], 'w') as outfile:
        for ex in examples:
            outfile.write(json.dumps(ex.to_dict()) + '\n')

    print(
        f"Extracted {len(examples)} self-feeding episodes out of "
        f"{len(episodes)} parlai episodes and wrote them to {config['outfile']}."
    )


if __name__ == '__main__':
    config = setup_args()
    assert config['infile'].endswith('.sliced')
    assert config['outfile'].endswith('.unfiltered')
    main(config)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://raw.githubusercontent.com/nlpdata/c3/master/data/c3-d-train.json',
        'train.json',
        'baf81f327dee84c6f451c9a4dd662e6193c67473b8791ffb72cce75cdb528f20',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/nlpdata/c3/master/data/c3-d-test.json',
        'test.json',
        'e9920491b31f9d00ecf31e51727b495dd6b0d05f4a96f273a343e81b6775a8f0',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/nlpdata/c3/master/data/c3-d-dev.json',
        'dev.json',
        '8c7054930a40aeb288ad7c51c42fa93d54aef678ccab29c75d46a7432f4f6278',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'C3')
    version = '1.0'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
import parlai.core.build_data as build_data
import codecs
import os

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/cornell_movie/cornell_movie_dialogs_corpus.tgz',
        'cornell_movie_dialogs_corpus.tgz',
        'ae77ab2e4743ce929087a4f529934059b920c4bdaa3143741b65b1e648ab45fd',
    )
]


def create_fb_format(lines_file, convo_file, outpath):
    print('[building fbformat]')
    with PathManager.open(
        os.path.join(outpath, 'train.txt'), 'w'
    ) as ftrain, PathManager.open(
        os.path.join(outpath, 'valid.txt'), 'w'
    ) as fvalid, PathManager.open(
        os.path.join(outpath, 'test.txt'), 'w'
    ) as ftest:
        lines = {}

        codecs.register_error('strict', codecs.ignore_errors)
        with codecs.open(lines_file, 'r') as f:
            for line in f:
                l = line.split(' +++$+++ ')
                lines[l[0]] = ' '.join(l[4:]).strip('\n').replace('\t', ' ')

        cnt = 0
        with codecs.open(convo_file, 'r') as f:
            for line in f:
                l = line.split(' ')
                convo = ' '.join(l[6:]).strip('\n').strip('[').strip(']')
                c = convo.replace("'", '').replace(' ', '').split(',')

                # forward conversation
                s = ''
                index = 0
                for i in range(0, len(c), 2):
                    index += 1
                    s += str(index) + ' ' + lines[c[i]]
                    if len(c) > i + 1:
                        s += '\t' + lines[c[i + 1]]
                    s += '\n'

                cnt = cnt + 1
                handle = ftrain
                if (cnt % 10) == 0:
                    handle = ftest
                if (cnt % 10) == 1:
                    handle = fvalid
                handle.write(s + '\n')


def build(opt):
    dpath = os.path.join(opt['datapath'], 'CornellMovie')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        dpext = os.path.join(dpath, 'cornell movie-dialogs corpus')
        create_fb_format(
            os.path.join(dpext, 'movie_lines.txt'),
            os.path.join(dpext, 'movie_conversations.txt'),
            dpath,
        )

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
PersonaChat task agents.

Persona can be 'none', 'self', 'other', or 'both'.
Format of persona can be 'original' or 'revised'.

This is specified in the following way:
--task personachat:{format}
...where {format} is one of...
- none
- self_original
- self_revised
- other_original
- other_revised
- both_original
- both_revised
"""


from parlai.core.teachers import FbDeprecatedDialogTeacher
from .build import build

import copy
import os


def _path(opt, persona):
    # Build the data if it doesn't exist.
    build(opt)
    dt = opt['datatype'].split(':')[0] + '_' + persona
    return os.path.join(opt['datapath'], 'Persona-Chat', 'personachat', dt + '.txt')


class NoneTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'none_original')
        super().__init__(opt, shared)


class SelfOriginalTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'self_original')
        super().__init__(opt, shared)


class SelfTeacher(SelfOriginalTeacher):
    pass


class SelfRevisedTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'self_revised')
        super().__init__(opt, shared)


class OtherOriginalTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'other_original')
        super().__init__(opt, shared)


class OtherTeacher(OtherOriginalTeacher):
    pass


class OtherRevisedTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'other_revised')
        super().__init__(opt, shared)


class BothOriginalTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'both_original')
        super().__init__(opt, shared)


class BothTeacher(BothOriginalTeacher):
    pass


class BothRevisedTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, 'both_revised')
        super().__init__(opt, shared)


class DefaultTeacher(SelfOriginalTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
import json
from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager

RESOURCES = [
    DownloadableFile(
        'https://computing.ece.vt.edu/~abhshkdz/data/visdial/visdial_0.9_train.zip',
        'visdial_0.9_train.zip',
        'a778d5d39d855b6194272f5800871a4a4b3673b00c9dc28d611443e7ca071290',
    ),
    DownloadableFile(
        'https://computing.ece.vt.edu/~abhshkdz/data/visdial/visdial_0.9_val.zip',
        'visdial_0.9_val.zip',
        '08f5ee1d0cb12620b311cb7efbce4bb43a586871f002adba541614877d6f3960',
    ),
]


def build(opt):
    version = 'v0.9'
    dpath = os.path.join(opt['datapath'], 'VisDial-v0.9')

    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')

        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        print('processing unpacked files')
        # Use 1000 examples from training set as validation.
        json1 = os.path.join(dpath, RESOURCES[0].file_name.rsplit('.', 1)[0] + '.json')
        with PathManager.open(json1) as t_json:
            train_data = json.load(t_json)

        valid_data = train_data.copy()
        valid_data['data'] = train_data['data'].copy()
        valid_data['data']['dialogs'] = []

        # Use constant stride to pick examples.
        num_valid = 1000
        total = len(train_data['data']['dialogs'])
        step = total // (num_valid - 1)
        for i in range(total - 1, 0, -step)[:num_valid]:
            valid_data['data']['dialogs'].append(train_data['data']['dialogs'][i])
            del train_data['data']['dialogs'][i]

        train_json = json1.rsplit('.', 1)[0] + '_train.json'
        valid_json = json1.rsplit('.', 1)[0] + '_valid.json'
        with PathManager.open(train_json, 'w') as t_out, PathManager.open(
            valid_json, 'w'
        ) as v_out:
            json.dump(train_data, t_out)
            json.dump(valid_data, v_out)
        PathManager.rm(json1)

        # Use validation data as test.
        json2 = os.path.join(dpath, RESOURCES[1].file_name.rsplit('.', 1)[0] + '.json')
        test_json = json2.rsplit('.', 1)[0] + '_test.json'
        build_data.move(json2, test_json)

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/v2_Questions_Train_mscoco.zip',
        'v2_Questions_Train_mscoco.zip',
        '05a64b6e2582d06d7585f5429674a9a33851878be1bff9f8668cdcf792df611e',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/v2_Questions_Val_mscoco.zip',
        'v2_Questions_Val_mscoco.zip',
        'e71f6c5c3e97a51d050f28243e262b28cd0c48d11a6b4632d769d30d3f93222a',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/v2_Questions_Test_mscoco.zip',
        'v2_Questions_Test_mscoco.zip',
        '982e2e687a86514b78ea83af356d151976c5e3fb4168a29ca543610574082ad7',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/v2_Annotations_Val_mscoco.zip',
        'v2_Annotations_Val_mscoco.zip',
        '0caae7c8d1dafd852727f5ac046bc1efca9b72026bd6ffa34fc489f3a7b3291e',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/v2_Annotations_Train_mscoco.zip',
        'v2_Annotations_Train_mscoco.zip',
        'fb101bcefe91422c543c2bb6d70af11eb3119d0ff745ae283d09acdf66250853',
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'VQA-v2')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        # An older version exists, so remove these outdated files.
        if build_data.built(dpath):
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Fastext vectors, e.g. use with filename "models:fasttext_vectors/wiki.en.vec".
"""

import os
import torchtext.vocab as vocab

URL = 'https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.en.vec'


def download(datapath):
    return vocab.Vectors(
        name='wiki.en.vec',
        url=URL,
        cache=os.path.join(datapath, 'models', 'fasttext_vectors'),
    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.utils.torch import concat_without_padding
from parlai.core.torch_ranker_agent import TorchRankerAgent
from parlai.zoo.bert.build import download

from .bert_dictionary import BertDictionaryAgent
from .helpers import BertWrapper, BertModel, add_common_args, surround, MODEL_PATH

import os


class CrossEncoderRankerAgent(TorchRankerAgent):
    """
    TorchRankerAgent implementation of the crossencoder.

    It is a standalone Agent. It might be called by the Both Encoder.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        add_common_args(parser)
        parser.set_defaults(
            encode_candidate_vecs=True,
            candidates='inline',
            dict_maxexs=0,  # skip building dictionary
        )
        return parser

    def __init__(self, opt, shared=None):
        # download pretrained models
        download(opt['datapath'])
        self.pretrained_path = os.path.join(
            opt['datapath'], 'models', 'bert_models', MODEL_PATH
        )

        super().__init__(opt, shared)
        # it's easier for now to use DataParallel when
        self.clip = -1
        self.NULL_IDX = self.dict.pad_idx
        self.START_IDX = self.dict.start_idx
        self.END_IDX = self.dict.end_idx

    def build_model(self):
        return BertWrapper(
            BertModel.from_pretrained(self.pretrained_path),
            1,
            add_transformer_layer=self.opt['add_transformer_layer'],
            layer_pulled=self.opt['pull_from_layer'],
            aggregation=self.opt['bert_aggregation'],
        )

    def score_candidates(self, batch, cand_vecs, cand_encs=None):
        # concatenate text and candidates (not so easy)
        # unpad and break
        nb_cands = cand_vecs.size()[1]
        size_batch = cand_vecs.size()[0]
        text_vec = batch.text_vec

        tokens_context = (
            text_vec.unsqueeze(1)
            .expand(-1, nb_cands, -1)
            .contiguous()
            .view(nb_cands * size_batch, -1)
        )

        # remove the start token ["CLS"] from candidates
        tokens_cands = cand_vecs.view(nb_cands * size_batch, -1)
        all_tokens, all_segments = concat_without_padding(
            tokens_context, tokens_cands, self.use_cuda, self.NULL_IDX
        )
        all_mask = all_tokens != self.NULL_IDX
        all_tokens *= all_mask.long()
        scores = self.model(all_tokens, all_segments, all_mask)
        return scores.view(size_batch, nb_cands)

    @staticmethod
    def dictionary_class():
        return BertDictionaryAgent

    def _set_text_vec(self, *args, **kwargs):
        obs = super()._set_text_vec(*args, **kwargs)
        # concatenate the [CLS] and [SEP] tokens
        if (
            obs is not None
            and 'text_vec' in obs
            and 'added_start_end_tokens' not in obs
        ):
            obs.force_set(
                'text_vec', surround(obs['text_vec'], self.START_IDX, self.END_IDX)
            )
            obs['added_start_end_tokens'] = True
        return obs



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Talk with a model using a web UI.
"""


from http.server import BaseHTTPRequestHandler, HTTPServer
from parlai.scripts.interactive import setup_args
from parlai.core.agents import create_agent
from parlai.core.worlds import create_task
from parlai.core.image_featurizers import ImageLoader
from typing import Dict, Any

import json
import cgi
import PIL.Image as Image
from base64 import b64decode
import io
import os

HOST_NAME = 'localhost'
PORT = 8080
SHARED: Dict[str, Any] = {}
IMAGE_LOADER = None
STYLE_SHEET = "https://cdnjs.cloudflare.com/ajax/libs/bulma/0.7.4/css/bulma.css"
FONT_AWESOME = "https://use.fontawesome.com/releases/v5.3.1/js/all.js"
WEB_HTML = """
<html>
    <link rel="stylesheet" href={} />
    <script defer src={}></script>
    <head><title> Interactive Run </title></head>
    <body>
        <div class="columns">
            <div class="column is-three-fifths is-offset-one-fifth">
              <section class="hero is-info is-large has-background-light has-text-grey-dark">
                <div id="parent" class="hero-body">
                    <article class="media" id="photo-info">
                      <figure class="media-left">
                        <span class="icon is-large">
                          <i class="fas fa-robot fas fa-2x"></i>
                        </span>
                      </figure>
                      <div class="media-content">
                        <div class="content">
                          <p>
                            <img id="preview" src="Examples.png"/ style="max-height:300px">
                          </p>
                        </div>
                      </div>
                      <div class="media-content">
                        <div class="content">
                          <p>
                            <strong>Model</strong>
                            <br>
                            Enter a message, and the model will respond interactively.
                          </p>
                        </div>
                      </div>
                    </article>
                </div>
                <div class="hero-foot column is-three-fifths is-offset-one-fifth">
                  <form id = "interact">
                      <div class="field is-grouped">
                        <p class="control is-expanded">
                          Upload an image:
                          <input class="input" type="file" id="userInImage" accept="image/*">
                        </p>
                            Choose a personality:
                            <select name="pers_list" form="interact" id="pers_select">
                            </select>
                        <p class="control">
                          <button id="respond" type="submit" class="button has-text-white-ter has-background-grey-dark">
                            Submit
                          </button>
                        </p>
                      </div>
                  </form>
                </div>
              </section>
            </div>
        </div>

        <script>
            var personalities = ["Adventurous", "Appreciative (Grateful)", "Articulate (Well-spoken, Expressive)", "Attractive", "Brilliant", "Calm", "Captivating", "Caring", "Charming", "Cheerful", "Clever", "Colorful (Full of Life, Interesting)", "Compassionate (Sympathetic, Warm)", "Confident", "Considerate", "Contemplative (Reflective, Thoughtful)", "Courageous", "Creative", "Cultured (Refined, Educated)", "Curious", "Daring", "Deep", "Dramatic", "Earnest (Enthusiastic)", "Elegant", "Eloquent (Well-spoken, Expressive)", "Empathetic", "Energetic", "Enthusiastic", "Exciting", "Extraordinary", "Freethinking", "Fun-loving", "Gentle", "Happy", "Honest", "Humble", "Humorous", "Idealistic", "Imaginative", "Insightful", "Intelligent", "Kind", "Knowledgeable", "Logical", "Meticulous (Precise, Thorough)", "Objective (Detached, Impartial)", "Observant", "Open", "Optimistic", "Passionate", "Patriotic", "Peaceful", "Perceptive", "Playful", "Practical", "Profound", "Rational", "Realistic", "Reflective", "Relaxed", "Respectful", "Romantic", "Rustic (Rural)", "Scholarly", "Sensitive", "Sentimental", "Serious", "Simple", "Sophisticated", "Spirited", "Spontaneous", "Stoic (Unemotional, Matter-of-fact)", "Suave (Charming, Smooth)", "Sweet", "Sympathetic", "Vivacious (Lively, Animated)", "Warm", "Wise", "Witty", "Youthful", "Absentminded", "Aggressive", "Amusing", "Artful", "Boyish", "Breezy (Relaxed, Informal)", "Businesslike", "Casual", "Cerebral (Intellectual, Logical)", "Complex", "Conservative (Traditional, Conventional)", "Contradictory", "Cute", "Dreamy", "Dry", "Emotional", "Enigmatic (Cryptic, Obscure)", "Formal", "Glamorous", "High-spirited", "Impersonal", "Intense", "Maternal (Mother-like)", "Mellow (Soothing, Sweet)", "Mystical", "Neutral", "Old-fashioned", "Ordinary", "Questioning", "Sarcastic", "Sensual", "Skeptical", "Solemn", "Stylish", "Tough", "Whimsical (Playful, Fanciful)", "Abrasive (Annoying, Irritating)", "Airy (Casual, Not Serious)", "Aloof (Detached, Distant)", "Angry", "Anxious", "Apathetic (Uncaring, Disinterested)", "Argumentative", "Arrogant", "Artificial", "Assertive", "Barbaric", "Bewildered (Astonished, Confused)", "Bizarre", "Bland", "Blunt", "Boisterous (Rowdy, Loud)", "Childish", "Coarse (Not Fine, Crass)", "Cold", "Conceited (Arrogant, Egotistical)", "Confused", "Contemptible (Despicable, Vile)", "Cowardly", "Crazy", "Critical", "Cruel", "Cynical (Doubtful, Skeptical)", "Destructive", "Devious", "Discouraging", "Disturbing", "Dull", "Egocentric (Self-centered)", "Envious", "Erratic", "Escapist (Dreamer, Seeks Distraction)", "Excitable", "Extravagant", "Extreme", "Fanatical", "Fanciful", "Fatalistic (Bleak, Gloomy)", "Fawning (Flattering, Deferential)", "Fearful", "Fickle (Changeable, Temperamental)", "Fiery", "Foolish", "Frightening", "Frivolous (Trivial, Silly)", "Gloomy", "Grand", "Grim", "Hateful", "Haughty (Arrogant, Snobbish)", "Hostile", "Irrational", "Irritable", "Lazy", "Malicious", "Melancholic", "Miserable", "Money-minded", "Monstrous", "Moody", "Morbid", "Narcissistic (Self-centered, Egotistical)", "Neurotic (Manic, Obsessive)", "Nihilistic", "Obnoxious", "Obsessive", "Odd", "Offhand", "Opinionated", "Outrageous", "Overimaginative", "Paranoid", "Passive", "Pompous (Self-important, Arrogant)", "Pretentious (Snobbish, Showy)", "Provocative", "Quirky", "Resentful", "Ridiculous", "Rigid", "Rowdy", "Scornful", "Shy", "Silly", "Stiff", "Stupid", "Tense", "Uncreative", "Unimaginative", "Unrealistic", "Vacuous (Empty, Unintelligent)", "Vague", "Wishful", "Zany"];

            function createChatRow(agent, text) {{
                var article = document.createElement("article");
                article.className = "media"

                var figure = document.createElement("figure");
                figure.className = "media-left";

                var span = document.createElement("span");
                span.className = "icon is-large";

                var icon = document.createElement("i");
                icon.className = "fas fas fa-2x" + (agent === "You" ? " fa-user " : agent === "Model" ? " fa-robot" : "");

                var media = document.createElement("div");
                media.className = "media-content";

                var content = document.createElement("div");
                content.className = "content";

                var para = document.createElement("p");
                para.id = "model-response";
                var paraText = document.createTextNode(text);

                var strong = document.createElement("strong");
                strong.innerHTML = agent;
                var br = document.createElement("br");

                para.appendChild(strong);
                para.appendChild(br);
                para.appendChild(paraText);
                content.appendChild(para);
                media.appendChild(content);

                span.appendChild(icon);
                figure.appendChild(span);

                article.appendChild(figure);
                article.appendChild(media);

                return article;
            }}
            function fetchResult(personality, image_data) {{
                var formData = new FormData();
                formData.append('personality', personality);
                formData.append('image', image_data);
                fetch('/interact', {{

                    method: 'POST',
                    body: formData
                }}).then(response=>response.json()).then(data=>{{
                    var parDiv = document.getElementById("parent");

                    // Change info for Model response
                    var oldResponse = document.getElementById("model-response");
                    if (oldResponse) {{
                        oldResponse.parentNode.remove(oldResponse);
                    }}
                    parDiv.append(createChatRow("Model", data.text));
                    window.scrollTo(0,document.body.scrollHeight);
                }});

            }}
            document.getElementById("interact").addEventListener("submit", function(event){{
                event.preventDefault()
                var personality = document.getElementById("pers_select").value;
                var img_input = document.getElementById("userInImage");
                var preview = document.getElementById("preview");
                if (img_input.files && img_input.files[0]) {{
                  var reader = new FileReader();
                  reader.onload = function (e) {{
                    preview.setAttribute('src', e.target.result);
                    fetchResult(personality, e.target.result);
                  }};
                  reader.readAsDataURL(img_input.files[0]);
                }} else {{
                  preview.setAttribute('src', 'Examples.png');
                }}
            }});
            var sel = document.getElementById("pers_select");
            for (var idx in personalities) {{
                personality = personalities[idx];
                var opt = document.createElement('option');
                opt.innerHTML = personality;
                opt.value = personality;
                sel.appendChild(opt);
            }}
        </script>

    </body>
</html>
"""  # noqa: E501


class MyHandler(BaseHTTPRequestHandler):
    """
    Interactive Handler.
    """

    def interactive_running(self, data):
        """
        Generate a model response.

        :param data:
            data to send to model

        :return:
            model act dictionary
        """
        reply = {}
        reply['text'] = data['personality'][0].decode()
        img_data = str(data['image'][0])
        _, encoded = img_data.split(',', 1)
        image = Image.open(io.BytesIO(b64decode(encoded))).convert('RGB')
        reply['image'] = SHARED['image_loader'].extract(image)
        SHARED['agent'].observe(reply)
        model_res = SHARED['agent'].act()
        return model_res

    def do_HEAD(self):
        """
        Handle headers.
        """
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def do_POST(self):
        """
        Handle POST.
        """
        if self.path != '/interact':
            return self.respond({'status': 500})
        ctype, pdict = cgi.parse_header(self.headers['content-type'])
        pdict['boundary'] = bytes(pdict['boundary'], "utf-8")
        postvars = cgi.parse_multipart(self.rfile, pdict)
        model_response = self.interactive_running(postvars)

        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        json_str = json.dumps(model_response)
        self.wfile.write(bytes(json_str, 'utf-8'))

    def do_GET(self):
        """
        Handle GET.
        """
        paths = {
            '/': {'status': 200},
            '/favicon.ico': {'status': 202},  # Need for chrome
        }
        if self.path in paths:
            self.respond(paths[self.path])
        else:
            self.respond({'status': 500})

    def handle_http(self, status_code, path, text=None):
        """
        Generate HTTP.
        """
        self.send_response(status_code)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        content = WEB_HTML.format(STYLE_SHEET, FONT_AWESOME)
        return bytes(content, 'UTF-8')

    def respond(self, opts):
        """
        Respond.
        """
        response = self.handle_http(opts['status'], self.path)
        self.wfile.write(response)


def setup_interactive():
    """
    Set up the interactive script.
    """
    parser = setup_args()
    opt = parser.parse_args()
    if not opt.get('model_file'):
        raise RuntimeError('Please specify a model file')
    if opt.get('fixed_cands_path') is None:
        opt['fixed_cands_path'] = os.path.join(
            '/'.join(opt.get('model_file').split('/')[:-1]), 'candidates.txt'
        )
    opt['task'] = 'parlai.agents.local_human.local_human:LocalHumanAgent'
    opt['image_mode'] = 'resnet152'
    SHARED['opt'] = opt
    SHARED['image_loader'] = ImageLoader(opt)

    # Create model and assign it to the specified task
    SHARED['agent'] = create_agent(opt, requireModelExists=True)
    SHARED['world'] = create_task(opt, SHARED['agent'])


if __name__ == '__main__':
    setup_interactive()
    server_class = HTTPServer
    Handler = MyHandler
    Handler.protocol_version = 'HTTP/1.0'
    httpd = server_class((HOST_NAME, PORT), Handler)
    print('http://{}:{}/'.format(HOST_NAME, PORT))

    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Basic example which allows local human keyboard input to talk to a trained model.

Note: this is identical to examples/interactive with the exception that we add
TransformerAgent command line args.
"""

import os
import random

from parlai.core.params import ParlaiParser
from parlai.core.agents import create_agent
from parlai.core.worlds import create_task
from parlai.agents.local_human.local_human import LocalHumanAgent
from parlai.tasks.self_feeding.build import build
from projects.self_feeding.self_feeding_agent import SelfFeedingAgent


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, True, 'Interactive chat with a model')
    parser.set_defaults(interactive_mode=True, task='interactive')
    LocalHumanAgent.add_cmdline_args(parser, partial_opt=None)
    SelfFeedingAgent.add_cmdline_args(parser, partial_opt=None)
    parser.set_defaults(history_size=2)
    return parser


def interactive(opt):
    opt['task'] = 'self_feeding'
    build(opt)
    opt['task'] = 'parlai.agents.local_human.local_human:LocalHumanAgent'
    cand_file = os.path.join(opt['datapath'], 'self_feeding/convai2_cands.txt')
    # Set values to override when the opt dict for the saved model is loaded
    opt['override'] = {
        'subtasks': ['dialog', 'satisfaction'],
        'interactive': True,
        'interactive_task': True,
        'prev_response_filter': True,
        'person_tokens': False,  # SelfFeedingAgent adds person_tokens on its own
        'partial_load': True,
        'history_size': 2,
        'eval_candidates': 'fixed',
        'encode_candidate_vecs': True,
        'fixed_candidates_path': cand_file,
        # Pull these from current opt dictionary
        'no_cuda': opt["no_cuda"],
        'fixed_candidate_vecs': opt['fixed_candidate_vecs'],
        'rating_frequency': opt['rating_frequency'],
        'rating_gap': opt['rating_gap'],
        'rating_threshold': opt['rating_threshold'],
        'request_feedback': opt['request_feedback'],
        'request_rating': opt['request_rating'],
    }

    # Create model and assign it to the specified task
    agent = create_agent(opt, requireModelExists=True)
    world = create_task(opt, agent)

    # Show some example dialogs:
    while True:
        world.parley()
        if world.epoch_done():
            print("EPOCH DONE")
            break


if __name__ == '__main__':
    random.seed(42)
    parser = setup_args()
    interactive(parser.parse_args())



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FbDeprecatedDialogTeacher, YamlTeacher
from parlai.utils.data import DatatypeHelper
from parlai.utils.misc import warn_once
from .build import build
from parlai.utils.strings import normalize_reply
import parlai.utils.logging as logging

import copy
import os

'''All teachers have a version with and without label candidates. Each teacher
defaults to using a dataset with label candidates. To use a dataset without
label candidates, specify this using the task flag:

--task convai2:{TEACHER_NAME}:no_cands

where TEACHER_NAME is None, SelfOriginal (Self), or SelfRevised.
'''


def _path(opt, persona, use_cands):
    # Build the data if it doesn't exist.
    build(opt)
    datatype = opt['datatype'].split(':')[0]
    if datatype == 'test':
        warn_once("WARNING: Test set not included. Setting datatype to valid.")
        datatype = 'valid'
    dt = datatype + '_' + persona
    cands = '' if use_cands else '_no_cands'
    return os.path.join(opt['datapath'], 'ConvAI2', dt + cands + '.txt')


class BothTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        try:
            cands = opt['task'].split(":")[2]
            use_cands = False if cands == 'no_cands' else True
        except Exception:
            use_cands = True
        opt['datafile'] = _path(opt, 'both_original', use_cands)
        super().__init__(opt, shared)


class NoneTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        try:
            cands = opt['task'].split(":")[2]
            use_cands = False if cands == 'no_cands' else True
        except Exception:
            use_cands = True
        opt['datafile'] = _path(opt, 'none_original', use_cands)
        super().__init__(opt, shared)


class SelfOriginalTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        try:
            cands = opt['task'].split(":")[2]
            use_cands = False if cands == 'no_cands' else True
        except Exception:
            use_cands = True
        opt['datafile'] = _path(opt, 'self_original', use_cands)
        super().__init__(opt, shared)


class SelfTeacher(SelfOriginalTeacher):
    pass


class SelfRevisedTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        try:
            cands = opt['task'].split(":")[2]
            use_cands = False if cands == 'no_cands' else True
        except Exception:
            use_cands = True
        opt['datafile'] = _path(opt, 'self_revised', use_cands)
        super().__init__(opt, shared)


class NormalizedTeacher(SelfOriginalTeacher):
    def normalize_replies(self, x):
        xs = x.split('\n')
        xs2 = []
        for x in xs:
            if 'your persona:' in x:
                # Normalize the sentence appearing after 'your persona:'
                x = x[len('your persona: ') :]
                x = normalize_reply(x)
                x = 'your persona: ' + x
            else:
                x = normalize_reply(x)

            xs2.append(x)
        return '\n'.join(xs2)

    def setup_data(self, path):
        logging.info(f"loading normalized fbdialog data: {path}")
        for (text, labels, reward, candidates), new_episode in super().setup_data(path):
            text = self.normalize_replies(text)
            labels = [self.normalize_replies(l) for l in labels]
            candidates = [self.normalize_replies(c) for c in candidates]
            yield (text, labels, reward, candidates), new_episode


class DefaultTeacher(SelfOriginalTeacher):
    pass


class InteractiveTeacher(SelfOriginalTeacher):
    # Dummy class to add arguments for interactive world.
    pass


class SelfchatTeacher(SelfOriginalTeacher):
    # Dummy class to add arguments for interactive world.
    pass


class SampleTeacher(YamlTeacher):
    """
    Loads the small sample of data created by the AutoTeacherTests.
    """

    def __init__(self, opt, shared=None):
        opt = opt.copy()
        fold = DatatypeHelper.fold(opt['datatype'])
        opt['datafile'] = os.path.join(
            os.path.dirname(__file__), f'test/convai2_{fold}.yml'
        )
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
import parlai.core.build_data as build_data
import os
import json

VERSION = '1.0'
RESOURCES = [
    DownloadableFile(
        'https://nlp.stanford.edu/data/coqa/coqa-train-v1.0.json',
        'coqa-train-v1.0.json',
        'b0fdb2bc1bd38dd3ca2ce5fa2ac3e02c6288ac914f241ac409a655ffb6619fa6',
        zipped=False,
    ),
    DownloadableFile(
        'https://nlp.stanford.edu/data/coqa/coqa-dev-v1.0.json',
        'coqa-dev-v1.0.json',
        'dfa367a9733ce53222918d0231d9b3bedc2b8ee831a2845f62dfc70701f2540a',
        zipped=False,
    ),
]


def make_parlai_format(outpath, dtype, data):
    print('building parlai:' + dtype)
    with PathManager.open(os.path.join(outpath, dtype + '.txt'), 'w') as fout:
        for each in data:
            output = []
            story = each['story'].replace('\n', '\\n')
            for question, ans in zip(each['questions'], each['answers']):
                question_txt = ''
                if question['turn_id'] == 1:
                    # include the context in the very first turn
                    question_txt = story + '\\n' + question['input_text']
                else:
                    question_txt = question['input_text']
                output.append(
                    'text:{question}\tlabels:{labels}'.format(
                        question=question_txt,
                        labels=ans['input_text'].replace('|', ' __PIPE__ '),
                    )
                )
                if question['turn_id'] < len(each['questions']):
                    output.append('\n')
            output.append('\t\tepisode_done:True\n')
            fout.write(''.join(output))


def build(opt):
    dpath = os.path.join(opt['datapath'], 'CoQA')
    version = VERSION

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        with PathManager.open(os.path.join(dpath, 'coqa-train-v1.0.json')) as f:
            data = json.load(f)['data']
            make_parlai_format(dpath, 'train', data)

        with PathManager.open(os.path.join(dpath, 'coqa-dev-v1.0.json')) as f:
            data = json.load(f)['data']
            make_parlai_format(dpath, 'valid', data)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Basic example which iterates through the tasks specified and load/extract the image
features.

For more options, check `parlai.core.image_featurizers`

## Examples

To extract the image feature of COCO images:

```shell
parlai extract_image_feature --task vqa_v1 --image-mode resnet152
```
"""
import copy
import tqdm

from parlai.core.params import ParlaiParser
from parlai.agents.repeat_label.repeat_label import RepeatLabelAgent
from parlai.core.worlds import create_task
import parlai.utils.logging as logging
from parlai.core.script import ParlaiScript, register_script


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, False, 'Load/extract image features')
    return parser


def extract_feats(opt):
    if isinstance(opt, ParlaiParser):
        logging.error('extract_feats should be passed opt not parser')
        opt = opt.parse_args()
    # Get command line arguments
    opt = copy.deepcopy(opt)
    dt = opt['datatype'].split(':')[0] + ':ordered'
    opt['datatype'] = dt
    opt['no_cuda'] = False
    opt['gpu'] = 0
    opt['num_epochs'] = 1
    opt['num_load_threads'] = 20
    opt.log()
    logging.info("Loading Images")
    # create repeat label agent and assign it to the specified task
    agent = RepeatLabelAgent(opt)
    world = create_task(opt, agent)

    total_exs = world.num_examples()
    pbar = tqdm.tqdm(unit='ex', total=total_exs)
    while not world.epoch_done():
        world.parley()
        pbar.update()
    pbar.close()

    logging.info("Finished extracting images")


@register_script('extract_image_feature', hidden=True)
class ExtractImgFeatures(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return extract_feats(self.opt)


if __name__ == '__main__':
    ExtractImgFeatures.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.teachers import FixedDialogTeacher
from parlai.utils.io import PathManager
from .build import build
import os
import json


class ELI5Teacher(FixedDialogTeacher):
    """
    ELI5 Teacher, taken from https://github.com/facebookresearch/ELI5.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        build(opt)
        self.id = 'eli5'
        self.messages = self.load_eli5(opt)
        self.reset()

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group('ELI5 Knowledge arguments')
        group.add_argument(
            '--knowledge',
            type='bool',
            default=True,
            help='Whether to include supporting document knowledge',
        )
        return parser

    def load_eli5(self, opt):
        """
        Load data based on data split.
        """
        dp = opt['datapath']
        dt = opt['datatype'].split(':')[0]
        eli_path = "eli5/processed_data/selected_15_1/explainlikeimfive_"
        fname = os.path.join(dp, eli_path + dt + ".json")
        if not PathManager.exists(fname):
            raise FileNotFoundError(
                f"{fname} not found. Please follow the instructions found at "
                "https://github.com/facebookresearch/ParlAI/tree/master/parlai/tasks/eli5/README.md"
                " to construct the dataset."
            )
        opt['datafile'] = fname
        with PathManager.open(fname) as json_file:
            data = json.load(json_file)
        ds = []
        for d in data:
            if self.opt['knowledge']:
                text = d['document'] + "\n" + d['question']
            else:
                text = d['question']
            act = {
                'id': 'eli5',
                'text': text,
                'labels': [d['answer']],
                'episode_done': True,
            }
            ds.append(act)
        return ds

    def get(self, episode_idx, entry_idx=0):
        return self.messages[episode_idx]

    def num_examples(self):
        return len(self.messages)

    def num_episodes(self):
        return len(self.messages)


class DefaultTeacher(ELI5Teacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/Questions_Train_mscoco.zip',
        'Questions_Train_mscoco.zip',
        'c3b2bb6155528eeae95e0a914af394d6f0d98f8f2b51012c44b27778e1a96707',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/Questions_Val_mscoco.zip',
        'Questions_Val_mscoco.zip',
        'e8839be5de2d711989bf0adc82e6717d1ce307d27c9b1dfb0abf413b79a5d4d0',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/Questions_Test_mscoco.zip',
        'Questions_Test_mscoco.zip',
        'bd080c297fc863bf8258caa4864d3b5afab29373375a6637f8546338291e28c0',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/Annotations_Val_mscoco.zip',
        'Annotations_Val_mscoco.zip',
        '29377c35186d90aeab3e61bdad890f51215d1f88b700bd22ef19004d73bf284f',
    ),
    DownloadableFile(
        'https://s3.amazonaws.com/cvmlp/vqa/mscoco/vqa/Annotations_Train_mscoco.zip',
        'Annotations_Train_mscoco.zip',
        'a5f5f97c162a4ad44896be08bac6deaa258aa3fec281afcc84fe85ae44cb1ebc',
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'VQA-v1')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data

RESOURCES = [
    DownloadableFile(
        '1WtbXCv3vPB5ql6w0FVDmAEMmWadbrCuG',
        'dialogue_nli.zip',
        '1164b0d9a0a1a6006891a6d4435a6a813464bc9b1e2f1ec5ce28c47267ad5e42',
        from_google=True,
    )
]


def build(opt):
    # get path to data directory
    dpath = os.path.join(opt['datapath'], 'dialogue_nli')
    # define version if any
    version = '1.0'

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from parlai.core.build_data import download_models


def download(datapath):
    opt = {'datapath': datapath}
    fnames = ['hh131k_hb60k_fb60k_st1k_v1.tar.gz']
    download_models(opt, fnames, 'self_feeding', version='v1.0', use_model_type=False)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Script for safety protected interaction between a local human keyboard input and a
trained model.
"""

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script
from parlai.core.agents import create_agent
from parlai.core.worlds import create_task
from parlai.agents.safe_local_human.safe_local_human import SafeLocalHumanAgent
import parlai.utils.logging as logging
import random


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, True, 'Like interactive, but adds a safety filter')
    parser.add_argument('-d', '--display-examples', type='bool', default=False)
    parser.add_argument(
        '--display-prettify',
        type='bool',
        default=False,
        help='Set to use a prettytable when displaying '
        'examples with text candidates',
    )
    parser.add_argument(
        '--display-add-fields',
        type=str,
        default='',
        help='Display these fields when verbose is off (e.g., "--display-add-fields label_candidates,beam_texts")',
    )
    parser.add_argument(
        '-it',
        '--interactive-task',
        type='bool',
        default=True,
        help='Create interactive version of task',
    )
    parser.set_defaults(interactive_mode=True, task='interactive')
    SafeLocalHumanAgent.add_cmdline_args(parser, partial_opt=None)
    return parser


def safe_interactive(opt):
    if isinstance(opt, ParlaiParser):
        logging.error('interactive should be passed opt not Parser')
        opt = opt.parse_args()

    # Create model and assign it to the specified task
    agent = create_agent(opt, requireModelExists=True)
    agent.opt.log()
    human_agent = SafeLocalHumanAgent(opt)
    world = create_task(opt, [human_agent, agent])

    # Interact until episode done
    while True:
        world.parley()
        bot_act = world.get_acts()[-1]
        if 'bot_offensive' in bot_act and bot_act['bot_offensive']:
            agent.reset()

        if opt.get('display_examples'):
            print('---')
            print(world.display())
        if world.epoch_done():
            logging.info('epoch done')
            break


@register_script('safe_interactive')
class SafeInteractive(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return safe_interactive(self.opt)


if __name__ == '__main__':
    random.seed(42)
    SafeInteractive.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
from parlai.utils.torch import neginf
from functools import lru_cache


def opt_to_kwargs(opt):
    """
    Get kwargs for seq2seq from opt.
    """
    kwargs = {}
    for k in ['memsize', 'time_features', 'position_encoding', 'hops']:
        if k in opt:
            kwargs[k] = opt[k]
    return kwargs


class MemNN(nn.Module):
    """
    Memory Network module.
    """

    def __init__(
        self,
        num_features,
        embedding_size,
        hops=1,
        memsize=32,
        time_features=False,
        position_encoding=False,
        dropout=0,
        padding_idx=0,
    ):
        """
        Initialize memnn model.

        See cmdline args in MemnnAgent for description of arguments.
        """
        super().__init__()

        # prepare features
        self.hops = hops

        def embedding(use_extra_feats=True):
            return Embed(
                num_features,
                embedding_size,
                position_encoding=position_encoding,
                padding_idx=padding_idx,
            )

        # TODO: add token dropout?
        # TODO: add dropout
        # self.dropout = nn.Dropout(dropout)
        # TODO: support more weight tying options?
        self.query_lt = embedding()
        self.in_memory_lt = embedding()
        self.out_memory_lt = embedding()
        self.answer_embedder = embedding()
        self.memory_hop = Hop(embedding_size)

    def forward(self, xs, mems, cands=None, pad_mask=None):
        """
        One forward step.

        :param xs:
            (bsz x seqlen) LongTensor queries to the model

        :param mems:
            (bsz x num_mems x seqlen) LongTensor memories

        :param cands:
            (num_cands x seqlen) or (bsz x num_cands x seqlen)
            LongTensor with candidates to rank
        :param pad_mask:
            (bsz x num_mems) optional mask indicating which tokens
            correspond to padding

        :returns:
            scores contains the model's predicted scores.
            if cand_params is None, the candidates are the vocabulary;
            otherwise, these scores are over the candidates provided.
            (bsz x num_cands)
        """
        state = self.query_lt(xs)
        if mems is not None:
            # no memories available, `nomemnn` mode just uses query/ans embs
            in_memory_embs = self.in_memory_lt(mems).transpose(1, 2)
            out_memory_embs = self.out_memory_lt(mems)

            for _ in range(self.hops):
                state = self.memory_hop(
                    state, in_memory_embs, out_memory_embs, pad_mask
                )

        if cands is not None:
            # embed candidates
            cand_embs = self.answer_embedder(cands)
        else:
            # rank all possible tokens
            cand_embs = self.answer_embedder.weight

        return state, cand_embs


class Embed(nn.Embedding):
    """
    Embed sequences for MemNN model.

    Applies Position Encoding if enabled and currently applies BOW sum.
    """

    def __init__(self, *args, position_encoding=False, reduction='mean', **kwargs):
        """
        Initialize custom Embedding layer.

        :param position_encoding:
            apply positional encoding transformation on input sequences

        :param reduction:
            reduction strategy to sequences, default 'mean'
        """
        self.position_encoding = position_encoding
        self.reduction = reduction
        super().__init__(*args, **kwargs)

    def _reduce(self, embs, input):
        # last dimension is embedding, do operation over dim before that
        if self.reduction == 'sum':
            return embs.sum(-2)
        elif self.reduction == 'mean':
            # this is more fair than mean(-2) since mean includes null tokens
            sum = embs.sum(-2)
            lens = (
                input.ne(self.padding_idx).sum(-1).unsqueeze(-1).float().clamp_(min=1)
            )
            return sum / lens
        else:
            raise RuntimeError(
                'reduction method {} not supported'.format(self.reduction)
            )

    def forward(self, input):
        """
        Return BOW embedding with PE reweighting if enabled.

        :param input:
            (bsz x seqlen) LongTensor

        :returns:
            (bsz x esz) FloatTensor
        """
        embs = super().forward(input)
        if self.position_encoding:
            if embs.dim() == 3:
                num_mems, seqlen, embdim = embs.size()
                pe = self.position_matrix(seqlen, embdim, embs.is_cuda)
                for i in range(num_mems):
                    embs[i] *= pe
            else:
                bsz, num_mems, seqlen, embdim = embs.size()
                pe = self.position_matrix(seqlen, embdim, embs.is_cuda)
                for i in range(num_mems):
                    embs[:, i] *= pe
        return self._reduce(embs, input)

    @staticmethod
    @lru_cache(maxsize=128)
    def position_matrix(J, d, use_cuda):
        """
        Build matrix of position encoding coeffiencents.

        See https://papers.nips.cc/paper/5846-end-to-end-memory-networks,
        section 4.1 Model Details: Sentence Representation.

        :param J:
            number of words in the sequence

        :param d:
            dimension of the embedding

        :returns:
            Position Encoding matrix
        """
        m = torch.Tensor(J, d)
        for k in range(1, d + 1):
            for j in range(1, J + 1):
                m[j - 1, k - 1] = (1 - j / J) - (k / d) * (1 - 2 * j / J)
        if use_cuda:
            m = m.cuda()
        return m


class Hop(nn.Module):
    """
    Memory Network hop outputs attention-weighted sum of memory embeddings.

    0) rotate the query embeddings 1) compute the dot product between the input vector
    and each memory vector 2) compute a softmax over the memory scores 3) compute the
    weighted sum of the memory embeddings using the probabilities 4) add the query
    embedding to the memory output and return the result
    """

    def __init__(self, embedding_size, rotate=True):
        """
        Initialize linear rotation.
        """
        super().__init__()
        if rotate:
            self.rotate = nn.Linear(embedding_size, embedding_size, bias=False)
        else:
            self.rotate = lambda x: x
        self.softmax = nn.Softmax(dim=1)

    def forward(self, query_embs, in_mem_embs, out_mem_embs, pad_mask):
        """
        Compute MemNN Hop step.

        :param query_embs:
            (bsz x esz) embedding of queries

        :param in_mem_embs:
            bsz list of (num_mems x esz) embedding of memories for activation

        :param out_mem_embs:
            bsz list of (num_mems x esz) embedding of memories for outputs

        :param pad_mask
            (bsz x num_mems) optional mask indicating which tokens correspond to
            padding

        :returns:
            (bsz x esz) output state
        """
        # rotate query embeddings
        attn = torch.bmm(query_embs.unsqueeze(1), in_mem_embs).squeeze(1)
        if pad_mask is not None:
            attn[pad_mask] = neginf(attn.dtype)
        probs = self.softmax(attn)
        memory_output = torch.bmm(probs.unsqueeze(1), out_mem_embs).squeeze(1)
        output = memory_output + self.rotate(query_embs)
        return output



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os


def build(opt):
    version = 'v6.0'
    dpath = os.path.join(opt['datapath'], 'rephrase_sentences')

    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        fnames = [
            'rephrase_sentences_train_0703.txt',
            'rephrase_sentences_valid_0703.txt',
            'rephrase_sentences_test_0703.txt',
            'choose_sentence_train.txt',
            'choose_sentence_valid.txt',
            'choose_sentence_test.txt',
        ]
        for fname in fnames:
            url = 'http://parl.ai/downloads/projects/rephrase_sentences/' + fname
            build_data.download(url, dpath, fname)

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This script is derived from parlai/core/scripts/eval_wordstat.py.

This script measures many different metrics of the text generated for the validation
set - including all the controllable attributes.
"""

from parlai.core.params import ParlaiParser
from parlai.core.dict import DictionaryAgent
from parlai.core.agents import create_agent
from parlai.core.worlds import create_task
from parlai.utils.misc import TimeLogger
from parlai.core.metrics import normalize_answer
from parlai.core.logs import TensorboardLogger
from controllable_seq2seq.controls import (
    ATTR2SENTSCOREFN,
    eval_attr,
    initialize_control_information,
)
from controllable_seq2seq.util import ConvAI2History
from collections import Counter
from parlai.utils.io import PathManager

import copy
import random
import json
import time
import os


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, True, 'compute statistics from model predictions')
    DictionaryAgent.add_cmdline_args(parser, partial_opt=None)

    # These defaults can be overriden by both .opt file and user's command line flags
    parser.add_argument('-ne', '--num-examples', type=int, default=-1)
    parser.add_argument('-ltim', '--log-every-n-secs', type=float, default=2)
    parser.add_argument(
        '-ed',
        '--external-dict',
        type=str,
        default=None,
        help='External dictionary for stat computation',
    )
    parser.add_argument(
        '-fb',
        '--freq-bins',
        type=str,
        default='0,100,1000,10000',
        help='Bins boundaries for rare words stat',
    )
    parser.add_argument(
        '-gr',
        '--gold-response',
        type=bool,
        default=False,
        help='Compute stats for gold response',
    )

    # These settings override .opt file but not user's command line flags
    parser.set_params(
        datatype='valid',
        task='projects.controllable_dialogue.tasks.agents',
        model='projects.controllable_dialogue.controllable_seq2seq.controllable_seq2seq:ControllableSeq2seqAgent',  # noqa: E501
        batchsize=64,
        beam_size=20,
        beam_min_n_best=10,
        use_reply='model',
    )
    TensorboardLogger.add_cmdline_args(parser, partial_opt=None)
    return parser


def get_word_stats(text, agent_dict, bins=(0, 100, 1000, 100000)):
    """
    Function which takes text sequence and dict, returns word freq and length
    statistics.

    :param sequence: text sequence
    :param agent_dict: can be external dict or dict from the model
    :param bins: list with range boundaries
    :return: freqs dictionary, num words, avg word length, avg char length
    """
    pred_list = agent_dict.tokenize(text)
    pred_freq = [agent_dict.freq[word] for word in pred_list]
    freqs = {i: 0 for i in bins}
    for f in pred_freq:
        for b in bins:
            if f <= b:
                freqs[b] += 1
                break

    wlength = len(pred_list)
    clength = len(text)  # including spaces
    return freqs, len(pred_freq), wlength, clength


def update_sent_attr_stats(sent_attrs, history, prediction):
    """
    Update the sent_attrs dict with the attributes of a prediction with given history.

    Inputs:
      sent_attrs: dictionary mapping each attr (a string) to a list of floats
        (the scores).
      history: a ConvAI2History
      prediction: string. the response text for which we measure sent attributes
    """
    for attr in sent_attrs.keys():
        attr_score = eval_attr(prediction, history, attr)
        sent_attrs[attr].append(attr_score)
    return sent_attrs


def eval_wordstat(opt):
    """
    Evaluates a model.

    :param opt: tells the evaluation function how to run
    """
    random.seed(42)

    # Setup control information
    initialize_control_information(opt)

    # Create model and assign it to the specified task
    agent = create_agent(opt, requireModelExists=True)
    world = create_task(opt, agent)

    if opt.get('external_dict'):
        print('[ Using external dictionary from: {} ]'.format(opt['external_dict']))
        dict_opt = copy.deepcopy(opt)
        dict_opt['dict_file'] = opt['external_dict']
        dictionary = DictionaryAgent(dict_opt)
    else:
        print('[ Using model bundled dictionary ]')
        dictionary = agent.dict

    batch_size = opt['batchsize']

    log_every_n_secs = opt.get('log_every_n_secs', -1)
    if log_every_n_secs <= 0:
        log_every_n_secs = float('inf')
    log_time = TimeLogger()

    data = {}  # This will be written to the output json file
    data['opt'] = agent.opt  # Save the opt to json

    # Determine the output filename
    if opt['gold_response']:  # Special output file for gold response
        model_dir, _ = os.path.split(opt.get('model_file'))
        outfile = os.path.join(model_dir, 'goldresponse')
        if opt['use_reply'] != 'label':
            raise ValueError(
                'You should set --use-reply label (not --use-reply model) '
                'when measuring goldresponse stats'
            )
    else:
        outfile = "%s.%s.%s.%s" % (
            opt.get('model_file'),
            opt.get('datatype'),
            "use%sreply" % agent.opt['use_reply'],
            "beam%i" % agent.opt['beam_size'],
        )
        if agent.opt['beam_size'] > 1:
            outfile += ".beamminnbest%i" % agent.opt['beam_min_n_best']
        if len(agent.control_settings) > 0:
            outfile += ".setcontrols:" + "_".join(
                [
                    "%s%s" % (c, str(agent.control_settings[c]['set_value']))
                    for c in sorted(agent.control_settings.keys())
                ]
            )
        if agent.opt['beam_reorder'] not in ['none', False]:
            outfile += ".beamreorder_%s" % agent.opt['beam_reorder']
        if len(agent.wd_features) > 0:
            sorted_bfw = sorted(
                list(zip(agent.wd_features, agent.wd_wts)), key=lambda x: x[0]
            )
            outfile += ".WDfeatures:" + "_".join(
                ["%s%s" % (f, str(w)) for f, w in sorted_bfw]
            )
    if opt['num_examples'] != -1:
        outfile += ".numex%i" % opt['num_examples']
    outfile += ".wordstats.json"
    print("\nOutfile: %s\n" % outfile)

    cnt = 0
    word_statistics = {
        'mean_wlength': [],  # list of length (in words) of utterances
        'mean_clength': [],  # list of length (in chars) of utterances
        'freqs_cnt': Counter(),  # Counter for word frequencies, bucketed
        'word_cnt': 0,  # total number of words in all utterances
        'pred_list': [],  # list of generated utterances after applying normalize_answer
        'pure_pred_list': [],  # list of generated utterances
        'context_list': [],  # list of text inputs (persona and conversation history)
    }
    bins = [int(i) for i in opt['freq_bins'].split(',')]

    # This dictionary records all the sentence-level controllable attributes
    # For each attribute, we have a list of all the values
    sent_attrs = {attr: [] for attr in ATTR2SENTSCOREFN.keys()}  # str to list of floats

    # histories will be a list of ConvAI2History objects
    histories = []

    def process_prediction(prediction, word_statistics):
        word_statistics['pred_list'].append(normalize_answer(prediction))
        freqs, _cnt, wlength, clength = get_word_stats(
            prediction, dictionary, bins=bins
        )
        word_statistics['word_cnt'] += _cnt
        word_statistics['mean_wlength'].append(wlength)
        word_statistics['mean_clength'].append(clength)
        word_statistics['freqs_cnt'] += Counter(freqs)
        return word_statistics

    t0 = time.time()
    while not world.epoch_done():
        world.parley()
        # orig eval_wordstat.py handles bsz=1 but for simplicity we assume bsz>1
        assert batch_size != 1
        for w in world.worlds:
            try:
                try:
                    response_act = w.acts[-1]
                    prediction = response_act['text']
                except KeyError:
                    continue
                if opt['gold_response']:
                    # If we're measuring gold response, use eval_label as prediction
                    prediction = w.acts[0]['eval_labels'][0]
                    response_act = {'text': prediction}
                word_statistics['context_list'].append(w.acts[0]['text'])
                word_statistics['pure_pred_list'].append(prediction)
            except IndexError:
                continue
            cnt += 1
            word_statistics = process_prediction(prediction, word_statistics)

            # Compute and record sentence-level attributes
            history = ConvAI2History(w.acts[0]['text'])
            histories.append(history)
            sent_attrs = update_sent_attr_stats(sent_attrs, history, prediction)

        # Periodically log some info
        if log_time.time() > log_every_n_secs:
            report = world.report()
            text, report = log_time.log(report['exs'], world.num_examples(), report)
            print(text)

        if opt['num_examples'] > 0 and cnt >= opt['num_examples']:
            break
    if world.epoch_done():
        print("EPOCH DONE")
    print("Time to process %i examples: %f seconds" % (cnt, time.time() - t0))

    # Compute percent unique
    # Note this is w.r.t. normalized pred_list not original pure_pred_list
    unique_list = []
    cntr = Counter(word_statistics['pred_list'])
    for k, v in cntr.items():
        if v == 1:
            unique_list.append(k)
    unique_percent = len(unique_list) / len(word_statistics['pred_list']) * 100

    # Print a final report
    report = world.report()
    if opt['gold_response']:
        report['ppl'] = 0.0  # For gold responses, overwrite the perplexity
    print(report)

    # Put all information in data dict
    data['unique_percent'] = unique_percent  # percent of all responses that are unique
    data['word_statistics'] = word_statistics  # word stats, as in orig eval_wordstat
    data['report'] = report  # the final report
    data['histories'] = [
        (hist.persona_lines, hist.partner_utts, hist.own_utts) for hist in histories
    ]  # history for each example
    data['sent_attrs'] = sent_attrs  # all sentence attribute values for responses

    # Write data to outfile
    print("Writing to %s..." % outfile)
    with PathManager.open(outfile, 'w') as f:
        json.dump(data, f)


if __name__ == '__main__':
    parser = setup_args()
    eval_wordstat(parser.parse_args())



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Utilities relating to frontend code.
"""

import os
import shutil
import subprocess


def build_task(task_directory: str):
    """
    Build the task with npm.
    """

    # Paths
    frontend_source_dir = os.path.join(task_directory, "webapp")
    frontend_build_dir = os.path.join(frontend_source_dir, "build")
    return_dir = os.getcwd()

    # Build the task
    os.chdir(frontend_source_dir)
    if os.path.exists(frontend_build_dir):
        shutil.rmtree(frontend_build_dir)
    packages_installed = subprocess.call(["npm", "install"])
    if packages_installed != 0:
        raise Exception(
            "please make sure npm is installed, otherwise view "
            "the above error for more info."
        )
    webpack_complete = subprocess.call(["npm", "run", "dev"])
    if webpack_complete != 0:
        raise RuntimeError(
            "Webpack appears to have failed to build your "
            "frontend. See the above error for more information."
        )
    os.chdir(return_dir)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import torch
import torch.nn as nn

from parlai.agents.transformer.modules import TransformerEncoder


class SelfFeedingModel(nn.Module):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        model = parser.add_argument_group('Self Feeding Model')

        model.add_argument(
            '-shl',
            '--sat-head-layers',
            type=int,
            default=1,
            help="The number of linear layers in the " "satisfaction task head",
        )
        model.add_argument(
            '-sfeeemb',
            '--share-fee-embeddings',
            type='bool',
            default=True,
            help="If True, the feedback task shares " "the dialog embeddings",
        )
        model.add_argument(
            '-sfeexenc',
            '--share-fee-x-encoder',
            type='bool',
            default=True,
            help="If True, the feedback task shares " "the dialog x encoder",
        )
        model.add_argument(
            '-sfeeyenc',
            '--share-fee-y-encoder',
            type='bool',
            default=True,
            help="If True, the feedback task shares " "the dialog y encoder",
        )
        model.add_argument(
            '-ssatemb',
            '--share-sat-embeddings',
            type='bool',
            default=False,
            help="If True, the satisfaction task shares " "the dialog embeddings",
        )
        model.add_argument(
            '-ssatenc',
            '--share-sat-encoder',
            type='bool',
            default=False,
            help="If True, the satisfaction task shares the dialog " "encoder",
        )
        return parser

    def __init__(self, opt, dictionary):
        super().__init__()
        self.opt = opt
        self.pad_idx = dictionary[dictionary.null_token]
        self.vocab_size = len(dictionary)

        # Build dialog
        self.dia_embeddings = self.init_embeddings()
        self.x_dia_encoder = self.build_encoder(opt, self.dia_embeddings)
        self.x_dia_head = nn.Dropout(p=0)

        self.y_dia_encoder = self.build_encoder(opt, self.dia_embeddings)
        self.y_dia_head = nn.Dropout(p=0)

        # Only build the parts of the network you will be using
        # This saves space (nbd) and prevents conflicts when loading
        # Build feedback
        if 'feedback' in self.opt['subtasks']:
            if self.opt['share_fee_embeddings']:
                self.fee_embeddings = self.dia_embeddings
            else:
                self.fee_embeddings = self.init_embeddings()
            if self.opt['share_fee_x_encoder']:
                self.x_fee_encoder = self.x_dia_encoder
            else:
                self.x_fee_encoder = self.build_encoder(opt, self.fee_embeddings)
            self.x_fee_head = nn.Dropout(p=0)

            if self.opt['share_fee_y_encoder']:
                self.y_fee_encoder = self.y_dia_encoder
            else:
                self.y_fee_encoder = self.build_encoder(opt, self.fee_embeddings)
            self.y_fee_head = nn.Dropout(p=0)

        # Build satisfaction
        if 'satisfaction' in self.opt['subtasks']:
            if self.opt['share_sat_embeddings']:
                self.sat_embeddings = self.dia_embeddings
            else:
                self.sat_embeddings = self.init_embeddings()
            if self.opt['share_sat_encoder']:
                self.x_sat_encoder = self.x_dia_encoder
            else:
                self.x_sat_encoder = self.build_encoder(opt, self.sat_embeddings)
            self.x_sat_head = self.build_head(
                opt, outdim=1, num_layers=self.opt['sat_head_layers']
            )

    def forward(self):
        raise NotImplementedError

    def score_dialog(self, x_vecs, y_vecs):
        x_enc = self.x_dia_head(self.x_dia_encoder(x_vecs))

        if y_vecs.dtype == torch.float32:
            # Assume candidates have already been encoded (e.g., in interactive mode)
            y_enc = y_vecs
        elif y_vecs.dtype == torch.int64:
            # Assume candidates have only been vectorized
            y_enc = self.encode_dia_y(y_vecs)
        else:
            raise Exception("Unsupported type for cands: {}".format(type(y_vecs)))

        return self.score_similarity(x_enc, y_enc)

    def encode_dia_y(self, y_vecs):
        """
        Encodes a tensor of vectorized candidates.

        :param y_vecs: a [bs, seq_len] or [bs, num_cands, seq_len](?) of vectorized
            candidates
        """
        if y_vecs.dim() == 2:
            y_enc = self.y_dia_head(self.y_dia_encoder(y_vecs))
        elif y_vecs.dim() == 3:
            oldshape = y_vecs.shape
            y_vecs = y_vecs.reshape(oldshape[0] * oldshape[1], oldshape[2])
            y_enc = self.y_dia_head(self.y_dia_encoder(y_vecs))
            y_enc = y_enc.reshape(oldshape[0], oldshape[1], -1)
        return y_enc

    def score_feedback(self, x_vecs, y_vecs):
        x_enc = self.x_fee_head(self.x_fee_encoder(x_vecs))
        y_enc = self.y_fee_head(self.y_fee_encoder(y_vecs))
        return self.score_similarity(x_enc, y_enc)

    def score_satisfaction(self, x_vecs):
        return torch.sigmoid(self.x_sat_head(self.x_sat_encoder(x_vecs))).squeeze(1)

    def score_similarity(self, context_h, cand_h):
        """
        Returns the dot product of encoded contexts and encoded candidates.
        """
        if self.opt['normalize_sent_emb']:
            context_h /= context_h.norm(2, dim=1, keepdim=True)
            cand_h /= cand_h.norm(2, dim=1, keepdim=True)

        if cand_h.dim() == 2:
            scores = torch.matmul(context_h, cand_h.t())
        elif cand_h.dim() == 3:
            scores = torch.bmm(context_h.unsqueeze(1), cand_h.transpose(1, 2)).squeeze(
                1
            )
        else:
            raise RuntimeError(
                'Unexpected candidate dimensions {}' ''.format(cand_h.dim())
            )

        return scores

    def init_embeddings(self):
        embeddings = nn.Embedding(
            self.vocab_size, self.opt['embedding_size'], padding_idx=self.pad_idx
        )
        nn.init.normal_(
            embeddings.weight, mean=0, std=self.opt['embedding_size'] ** -0.5
        )
        nn.init.constant_(embeddings.weight[self.pad_idx], 0)
        return embeddings

    def build_encoder(self, opt, embeddings):
        return TransformerEncoder(
            n_heads=opt['n_heads'],
            n_layers=opt['n_layers'],
            embedding_size=opt['embedding_size'],
            ffn_size=opt['ffn_size'],
            vocabulary_size=self.vocab_size,
            embedding=embeddings,
            attention_dropout=opt['attention_dropout'],
            relu_dropout=opt['relu_dropout'],
            padding_idx=self.pad_idx,
            learn_positional_embeddings=opt.get('learn_positional_embeddings', False),
            embeddings_scale=opt['embeddings_scale'],
        )

    def build_head(self, opt, outdim=1, num_layers=1):
        dim = self.opt['embedding_size']
        modules = []
        for _ in range(num_layers - 1):
            modules.append(nn.Linear(dim, dim))
            modules.append(nn.ReLU())
        modules.append(nn.Linear(dim, outdim))
        return nn.Sequential(*modules)


class Identity(nn.Module):
    def forward(self, x):
        return x



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
The general ParlAI Script interface.

An abstract class to help standardize the call to ParlAI scripts, enabling them to be
completed easily.

Also contains helper classes for loading scripts, etc.
"""

import sys
import io
import argparse
from typing import List, Optional, Dict, Any
import parlai
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser, CustomHelpFormatter
from abc import abstractmethod
import importlib
import pkgutil
import parlai.scripts
import parlai.utils.logging as logging
from parlai.core.loader import register_script, SCRIPT_REGISTRY  # noqa: F401


def setup_script_registry():
    """
    Loads the scripts so that @register_script is hit for all.
    """
    for module in pkgutil.iter_modules(parlai.scripts.__path__, 'parlai.scripts.'):
        importlib.import_module(module.name)


class ParlaiScript(object):
    """
    A ParlAI script is a standardized form of access.
    """

    parser: ParlaiParser

    @classmethod
    @abstractmethod
    def setup_args(cls) -> ParlaiParser:
        """
        Create the parser with args.
        """
        # we want to later deprecate this for add_cmdline_args

    def __init__(self, opt: Opt):
        self.opt = opt

    @abstractmethod
    def run(self):
        """
        The main method.

        Must be implemented by the script writer.
        """
        raise NotImplementedError()

    @classmethod
    def _run_kwargs(cls, kwargs: Dict[str, Any]):
        """
        Construct and run the script using kwargs, pseudo-parsing them.
        """
        parser = cls.setup_args()
        opt = parser.parse_kwargs(**kwargs)
        return cls._run_from_parser_and_opt(opt, parser)

    @classmethod
    def _run_args(cls, args: Optional[List[str]] = None):
        """
        Construct and run the script using args, defaulting to getting from CLI.
        """
        parser = cls.setup_args()
        opt = parser.parse_args(args=args)
        return cls._run_from_parser_and_opt(opt, parser)

    @classmethod
    def _run_from_parser_and_opt(cls, opt: Opt, parser: ParlaiParser):
        script = cls(opt)
        script.parser = parser
        return script.run()

    @classmethod
    def main(cls, *args, **kwargs):
        """
        Run the program, possibly with some given args.

        You may provide command line args in the form of strings, or
        options. For example:

        >>> MyScript.main(['--task', 'convai2'])
        >>> MyScript.main(task='convai2')

        You may not combine both args and kwargs.
        """
        assert not (bool(args) and bool(kwargs))
        if args:
            return cls._run_args(args)
        elif kwargs:
            return cls._run_kwargs(kwargs)
        else:
            return cls._run_args(None)

    @classmethod
    def help(cls, **kwargs):
        f = io.StringIO()
        parser = cls.setup_args()
        parser.prog = cls.__name__
        parser.add_extra_args(parser._kwargs_to_str_args(**kwargs))
        parser.print_help(f)
        return f.getvalue()


class _SupercommandParser(ParlaiParser):
    """
    Specialty ParlAI parser used for the supercommand.

    Contains some special behavior.
    """

    def __init__(self, *args, **kwargs):
        # used to target help messages more correctly, see GH #3182
        self._help_subparser = None

        from parlai.utils.strings import colorize

        logo = ""
        logo += colorize('       _', 'red') + "\n"
        logo += colorize('      /', 'red') + colorize('"', 'brightblack')
        logo += colorize(")", "yellow") + "\n"
        logo += colorize('     //', 'red') + colorize(')', 'yellow') + '\n'
        logo += colorize('  ==', 'green')
        logo += colorize("/", 'blue') + colorize('/', 'red') + colorize("'", 'yellow')
        logo += colorize("===", 'green') + " ParlAI\n"
        logo += colorize("   /", 'blue')
        kwargs['description'] = logo
        return super().__init__(*args, **kwargs)

    def add_extra_args(self, args):
        sa = [a for a in self._actions if isinstance(a, argparse._SubParsersAction)]
        assert len(sa) == 1
        sa = sa[0]
        if args is None:
            args = sys.argv[1:]
        if args and args[0] in sa.choices:
            # if at all possible, try to use the actual subcommand only for
            # the add extra args, to prevent parse errors in other commands.
            sa.choices[args[0]].add_extra_args(args)
        else:
            for _, v in sa.choices.items():
                v.add_extra_args(args)

    def parse_known_args(self, args=None, namespace=None, nohelp=False):
        known, unused = super().parse_known_args(args, namespace, nohelp)
        if hasattr(known, '_subparser'):
            # keep this around to keep the print message more in tune
            self._help_subparser = known._subparser
        return known, unused

    def print_help(self):
        """
        Print help, possibly deferring to the appropriate subcommand.
        """
        if self._help_subparser:
            self._help_subparser.print_help()
        else:
            return super().print_help()

    def add_subparsers(self, **kwargs):
        return super().add_subparsers(**kwargs)

    def _unsuppress_hidden(self):
        """
        Restore the help messages of hidden commands.
        """

        spa = [a for a in self._actions if isinstance(a, argparse._SubParsersAction)]
        assert len(spa) == 1
        spa = spa[0]
        for choices_action in spa._choices_actions:
            dest = choices_action.dest
            if choices_action.help == argparse.SUPPRESS:
                choices_action.help = spa.choices[dest].description

    def print_helpall(self):
        self._unsuppress_hidden()
        self.print_help()


class _SubcommandParser(ParlaiParser):
    """
    ParlaiParser which always sets add_parlai_args and add_model_args to False.

    Used in the superscript to initialize just the args for that command.
    """

    def __init__(self, **kwargs):
        kwargs['add_parlai_args'] = False
        kwargs['add_model_args'] = False
        assert 'description' in kwargs, 'Must supply description'
        return super().__init__(**kwargs)

    def parse_known_args(self, args=None, namespace=None, nohelp=False):
        if not nohelp:
            self.add_extra_args(args)
        return super().parse_known_args(args, namespace, nohelp)


def _SuperscriptHelpFormatter(**kwargs):
    kwargs['width'] = 100
    kwargs['max_help_position'] = 9999

    return CustomHelpFormatter(**kwargs)


def superscript_main(args=None):
    """
    Superscript is a loader for all the other scripts.
    """
    setup_script_registry()

    parser = _SupercommandParser(
        False, False, formatter_class=_SuperscriptHelpFormatter
    )
    parser.add_argument(
        '--helpall',
        action='helpall',
        help='List all commands, including advanced ones.',
    )
    parser.add_argument(
        '--version',
        action='version',
        version=get_version_string(),
        help='Prints version info and exit.',
    )
    parser.set_defaults(super_command=None)
    subparsers = parser.add_subparsers(
        parser_class=_SubcommandParser, title="Commands", metavar="COMMAND"
    )
    hparser = subparsers.add_parser(
        'help',
        aliases=['h'],
        help=argparse.SUPPRESS,
        description='List the main commands.',
    )
    hparser.set_defaults(super_command='help')
    hparser = subparsers.add_parser(
        'helpall',
        help=argparse.SUPPRESS,
        description='List all commands, including advanced ones.',
    )
    hparser.set_defaults(super_command='helpall')

    # build the supercommand
    for script_name, registration in SCRIPT_REGISTRY.items():
        logging.verbose(f"Discovered command {script_name}")
        script_parser = registration.klass.setup_args()
        if script_parser is None:
            # user didn't bother defining command line args. let's just fill
            # in for them
            script_parser = ParlaiParser(False, False)
        help_ = argparse.SUPPRESS if registration.hidden else script_parser.description
        subparser = subparsers.add_parser(
            script_name,
            aliases=registration.aliases,
            help=help_,
            description=script_parser.description,
            formatter_class=CustomHelpFormatter,
        )
        subparser.set_defaults(
            # carries the name of the full command so we know what to execute
            super_command=script_name,
            # used in ParlAI parser to find CLI options set by user
            _subparser=subparser,
        )
        subparser.set_defaults(**script_parser._defaults)
        for action in script_parser._actions:
            subparser._add_action(action)
        for action_group in script_parser._action_groups:
            subparser._action_groups.append(action_group)

    try:
        import argcomplete

        argcomplete.autocomplete(parser)
    except ModuleNotFoundError:
        pass

    opt = parser.parse_args(args)
    cmd = opt.pop('super_command')
    if cmd == 'helpall':
        parser.print_helpall()
    elif cmd == 'versioninfo':
        exit(0)
    elif cmd == 'help' or cmd is None:
        parser.print_help()
    elif cmd is not None:
        return SCRIPT_REGISTRY[cmd].klass._run_from_parser_and_opt(opt, parser)


def get_version_string() -> str:
    return f"ParlAI version {parlai.__version__}"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.tasks.self_chat.worlds import SelfChatWorld as SelfChatBaseWorld


class SelfChatWorld(SelfChatBaseWorld):
    def get_contexts(self):
        return None



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BART: Denoising Sequence-to-Sequence Pre-training for
Natural Language Generation, Translation, and Comprehension

See https://arxiv.org/abs/1910.13461.

The BART agent can be instantiated as simply `-m bart`,
however it is recommended to specify `--init-model zoo:bart/bart_large/model`
or `-mf zoo:bart/bart_large/model` to ensure correct dictionaries are saved.
"""
import os
import torch
from typing import Optional, Dict, Any

from parlai.agents.bart.convert_fairseq_to_parlai import ConversionScript
from parlai.agents.bart.modules import BartModel
from parlai.agents.transformer.transformer import TransformerGeneratorAgent
from parlai.core.agents import compare_init_model_opts
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import History
from parlai.utils.typing import TShared
from parlai.utils.io import PathManager
from parlai.zoo.bart.build import download, CONVERSION_ARGS, BART_ARGS


class BartAgent(TransformerGeneratorAgent):
    """
    BART Agent.

    Relies on the BART model implemented in fairseq.

    If you have a fine-tuned BART model from fairseq, you can specify the
    `--init-fairseq-model` arg, which will convert your fine-tuned model
    to a ParlAI model.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Override to add init-fairseq-model arg.
        """
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        group = parser.add_argument_group('Bart Args')
        group.add_argument(
            '--init-fairseq-model',
            type=str,
            default=None,
            help='fairseq checkpoint for bart',
        )
        group.add_argument(
            '--output-conversion-path',
            type=str,
            default=None,
            help='where to save fairseq conversion',
        )
        parser.set_defaults(dict_tokenizer='gpt2')
        parser.set_defaults(**BART_ARGS)
        return parser

    def __init__(self, opt: Opt, shared: TShared = None):
        if not shared:
            opt = self._initialize_bart(opt)
        super().__init__(opt, shared)

    def _initialize_bart(self, opt: Opt) -> Opt:
        """
        Download and convert BART pre-trained models.

        Additionally, convert `init-fairseq-model` if necessary.

        :param opt:
            ParlAI-parsed options

        :return opt:
            return opt with BART-specific args.
        """
        init_model, _ = self._get_init_model(opt, None)
        if not opt.get('converting') and (
            init_model is None or not PathManager.exists(init_model)
        ):
            download(opt['datapath'])
            opt['init_model'] = os.path.join(
                opt['datapath'], 'models/bart/bart_large/model'
            )
        if opt.get('init_fairseq_model'):
            opt = self._convert_model(opt)

        compare_init_model_opts(opt, opt)
        return opt

    def _get_conversion_args(self, opt: Opt) -> Dict[str, Any]:
        """
        Get args for fairseq model conversion.

        :param opt:
            ParlAI Opt

        :return args:
            returns dictionary of args to send to conversion script.
        """
        model_name = os.path.split(opt['init_fairseq_model'])[-1]
        args = CONVERSION_ARGS.copy()

        args['input'] = [opt['init_fairseq_model']]
        if opt.get('model_file') and not os.path.exists(opt['model_file']):
            args['output'] = opt['model_file']
        elif opt.get('output_conversion_path'):
            args['output'] = opt['output_conversion_path']
        else:
            args['output'] = os.path.join(
                opt['datapath'], 'models/converted_fairseq_models/', model_name
            )

        return args

    def _convert_model(self, opt: Opt) -> Opt:
        """
        Convert fairseq init model to ParlAI Model.

        :param opt:
            options

        :return opt:
            return opt with new init_model path
        """
        args = self._get_conversion_args(opt)
        ConversionScript.main(**args)
        opt['init_model'] = args['output']
        return opt

    def build_model(self) -> BartModel:
        """
        Build and return model.
        """
        model = BartModel(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model

    def _set_text_vec(
        self, obs: Message, history: History, truncate: Optional[int]
    ) -> Message:
        """
        Override to prepend start token and append end token.
        """
        obs = super()._set_text_vec(obs, history, truncate)
        if 'text' not in obs or 'text_vec' not in obs:
            return obs
        vec = obs['text_vec']
        if truncate is not None:
            vec = torch.LongTensor(  # type: ignore
                self._check_truncate(obs['text_vec'], truncate - 2, True)
            )
        obs.force_set(
            'text_vec', self._add_start_end_tokens(vec, add_start=True, add_end=True)
        )
        return obs

    def _get_initial_decoder_input(
        self, bsz: int, beam_size: int, dev: torch.device
    ) -> torch.LongTensor:
        """
        Override to seed decoder with EOS BOS token.
        """
        return (
            torch.LongTensor([self.END_IDX, self.START_IDX])  # type: ignore
            .expand(bsz * beam_size, 2)
            .to(dev)
        )
